//---------------------------------------------------------------------------

#include "ApplicationIo.h"
#include <Malibu_Mb.h>
#include <StringSupport_Mb.h>
#include <SystemSupport_Mb.h>
//#include <CommandLine_Mb.h>
#include <sstream>

using namespace Innovative;

//==============================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application Io Class
//==============================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    :  UI(ui)
{
    Timer.Interval(300);
	Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);

    Baseboard.Eeprom().OnJtagLoadProgress.SetEvent(this, &ApplicationIo::HandleLoadProgress);
    Baseboard.Eeprom().OnJtagLoadComplete.SetEvent(this, &ApplicationIo::HandleLoadComplete);
    Baseboard.Eeprom().OnJtagLoadMessage.SetEvent(this, &ApplicationIo::HandleLoadMessage);

    // Parse command-line
//	std::auto_ptr<CommandLineArguments> args = GetCommandLineArguments();
//    CommandLine line(static_cast<int>(args->size()), &args->begin()[0]);
//    Settings.Debug = line.search("-debug");
}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Open() -- Open target
//---------------------------------------------------------------------------

void ApplicationIo::Open(int target)
{
    try
        {
#ifndef NO_BUSMASTER
        Baseboard.BusMasterSize(1<<20);
#endif        
        Baseboard.Target(target);
        Baseboard.Open();
        UI->Log("Target driver opened ok");
        DisplayLogicVersion();
        }
    catch(...)
        {
        UI->Log("Could not open specified target");
        }
}

//------------------------------------------------------------------------------
//  ApplicationIo::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  ApplicationIo::DisplayLogicVersion()
{
    char pcb = 'A' + static_cast<char>(Baseboard.PciLogicVersion().PciLogicPcb());

    std::stringstream msg;
    msg << std::hex << "Logic Version: " << Baseboard.PciLogicVersion().PciLogicRevision()
        << ", Family: " << Baseboard.PciLogicVersion().PciLogicFamily()
        << ", PCB Revision: " << pcb
        << ", Type: " << Baseboard.PciLogicVersion().PciLogicType();
    UI->OnRevision(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::Close() -- Close target
//---------------------------------------------------------------------------

void ApplicationIo::Close()
{
    Baseboard.Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Load() -- Load logic from file
//---------------------------------------------------------------------------

void  ApplicationIo::Load()
{
    Watch.Start();
    std::stringstream msg;
    msg << "Elapsed: " << FloatToString(Watch.Elapsed(), 1) << " secs";
    UI->OnElapsed(msg.str());

    UI->Log("-------------------------------------------------------------");
    UI->Log("Programming Eeprom...");

    Timer.Enabled(true);

	if (Settings.LogFile)
        {
        std::string file = Settings.Path() + std::string("EEProm.log");
        Log.open(file.c_str());
        }

    return Baseboard.Eeprom().LoadJtag(Settings.XsvfFile, Settings.DebugLevel);
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- return # of boards found on bus
//---------------------------------------------------------------------------

unsigned int ApplicationIo::BoardCount()
{
    return static_cast<unsigned int>(Baseboard.BoardCount());
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardName() --  Return board name string
//---------------------------------------------------------------------------

std::string ApplicationIo::BoardName()
{
    return PmcDriver;
}

//---------------------------------------------------------------------------
//  ApplicationIo::LogicVersion() --  Return reference to EEPROM version object
//---------------------------------------------------------------------------

Innovative::PciLogicVersionInfoIntf & ApplicationIo::LogicVersion()
{
    return Baseboard.PciLogicVersion();
}

//---------------------------------------------------------------------------
//  ApplicationIo::PeekPort() --
//---------------------------------------------------------------------------

ii32 ApplicationIo::PeekPort(int offset)
{
    return Baseboard.PeekPort(offset);
}

//---------------------------------------------------------------------------
//  ApplicationIo::PokePort() --
//---------------------------------------------------------------------------

void ApplicationIo::PokePort(int offset, ii32 value)
{
    Baseboard.PokePort(offset, value);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadProgress() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadProgress(ProcessProgressEvent & event)
{
    UI->OnProgress(event.Percent);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadMessage() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadMessage(ProcessStatusEvent & event)
{
    std::string msg = Trim(event.Message);

	if (Settings.LogFile)
        Log << msg << std::endl;

    if (Settings.LogMessages)
        UI->OnLoadMessage(msg);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadComplete() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadComplete(ProcessCompletionEvent & event)
{
	if (Settings.LogFile)
        Log.close();

    UI->OnLoadComplete(event.Result);

    Timer.Enabled(false);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*event*/)
{
    std::stringstream msg;
    msg << "Elapsed: " << FloatToString(Watch.Elapsed(), 1) << " secs";
    UI->OnElapsed(msg.str());
}

//==============================================================================
//  CLASS ApplicationSettings - Remember Settings between runs
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class ApplicationSettings --  load from INI file
//------------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings()
{
	Application App;
    FPath = App.Path();
	std::string file = Path() + std::string("Settings.ini");
	Ini = new IniFile(file);

    Target  = Ini->Read("Target", 0);
    DebugLevel  = Ini->Read("DebugLevel", 1);
    LogFile  = Ini->Read("LogFile", false);
    LogMessages  = Ini->Read("LogMessages", false);
    XsvfFile = Ini->Read("XsvfFile", std::string(""));
}

//------------------------------------------------------------------------------
//  destructor for class ApplicationSettings --  save to INI file
//------------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    Ini->Write("Target",  Target);
    Ini->Write("DebugLevel",  DebugLevel);
    Ini->Write("LogFile",  LogFile);
    Ini->Write("LogMessages",  LogMessages);
    Ini->Write("XsvfFile",  XsvfFile);

	delete Ini;
	Ini = 0;
}


