//---------------------------------------------------------------------------

#include "ApplicationIo.h"
#include <Malibu_Mb.h>
#include <StringSupport_Mb.h>
#include <SystemSupport_Mb.h>
#include <sstream>

using namespace Innovative;
using namespace std;

//==============================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application Io Class
//==============================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    :  UI(ui), FOpen(false)
{
    Timer.Interval(300);
    Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);
}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Open() -- Open target
//---------------------------------------------------------------------------

void ApplicationIo::Open(int target)
{
    ostringstream msg;

    try
        {
        #ifndef NO_BUSMASTER
        // Minimal BM allocation
        Baseboard.BusMasterSize(0x1000);
        #endif
        Baseboard.Target(target);
        Baseboard.Open();
        FOpen = true;
        msg << "Successfully opened target " << target;
        Led(false);
        }
    catch(...)
        {
        msg << "Could not open target " << target;
        FOpen = false;
        }

    UI->Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::Close() -- Close target
//---------------------------------------------------------------------------

void ApplicationIo::Close()
{
    Baseboard.Close();
    FOpen = false;

   UI->Log("Closed target");
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- return # of boards found on bus
//---------------------------------------------------------------------------

unsigned int ApplicationIo::BoardCount()
{
    unsigned int count = static_cast<unsigned int>(Baseboard.BoardCount());

    ostringstream msg;
    msg << count << " targets detected";
    UI->Log(msg.str());

    return count;
}

//---------------------------------------------------------------------------
//  ApplicationIo::LicenseCheck() -- Check WD License
//---------------------------------------------------------------------------

void ApplicationIo::LicenseCheck()
{
	unsigned int status(0);
	bool result = Baseboard.LicenseCheck(status);
	if (result==false)
		{
		std::stringstream msg;
		msg << "WinDriver License FAILURE: code=" << hex << status;
		UI->Log(msg.str());
		}
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardName() --  Return board name string
//---------------------------------------------------------------------------

std::string ApplicationIo::BoardName()
{
    return "X5 XMC Module";
}

//---------------------------------------------------------------------------
//  ApplicationIo::Led() --  Set LED
//---------------------------------------------------------------------------

void  ApplicationIo::Led(bool state)
{
    Baseboard.Led(state);

    if (state)
        UI->Log("LED on");
    else
        UI->Log("LED off");
}

//---------------------------------------------------------------------------
//  ApplicationIo::ToggleLed() --  Toggle LED
//---------------------------------------------------------------------------

void  ApplicationIo::ToggleLed()
{
    Baseboard.Led(!Baseboard.Led());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*event*/)
{
    ToggleLed();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Blink() --  Set blink status
//---------------------------------------------------------------------------

void ApplicationIo::Blink(bool state)
{
    Timer.Enabled(state);
    if (state)
        UI->Log("Blinking enabled");
    else
        UI->Log("Blinking disabled");
}

//---------------------------------------------------------------------------
//  ApplicationIo::Blink() --  Query blink status
//---------------------------------------------------------------------------

bool ApplicationIo::Blink() const
{
    return Timer.Enabled();
}

//==============================================================================
//  CLASS ApplicationSettings - Remember Settings between runs
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class ApplicationSettings --  load from INI file
//------------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings()
{
    Application App;
    std::string file = App.Path() + std::string("Settings.ini");
    Ini = new IniFile(file);

    Target  = Ini->Read("Target", 0);
}

//------------------------------------------------------------------------------
//  destructor for class ApplicationSettings --  save to INI file
//------------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    Ini->Write("Target",  Target);

    delete Ini;
    Ini = 0;
}


