//---------------------------------------------------------------------------

#include "ApplicationIo.h"
#include <Malibu_Mb.h>
#include <StringSupport_Mb.h>
#include <SystemSupport_Mb.h>
#include <CommandLine_Mb.h>
#include <sstream>

using namespace Innovative;

//==============================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application Io Class
//==============================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    :  UI(ui), FOpen(false)
{
    Baseboard.Logic().OnFpgaParseProgress.SetEvent(this, &ApplicationIo::HandleParseProgress);
    Baseboard.Logic().OnFpgaParseComplete.SetEvent(this, &ApplicationIo::HandleParseComplete);
    Baseboard.Logic().OnFpgaParseMessage.SetEvent(this, &ApplicationIo::HandleLoadError);
    Baseboard.Logic().OnFpgaLoadProgress.SetEvent(this, &ApplicationIo::HandleLoadProgress);
    Baseboard.Logic().OnFpgaLoadComplete.SetEvent(this, &ApplicationIo::HandleLoadComplete);
    Baseboard.Logic().OnFpgaLoadMessage.SetEvent(this, &ApplicationIo::HandleLoadError);

	Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);
	Timer.OnElapsed.Thunk();

    // Parse command-line
    CommandLineArguments args = GetCommandLineArguments();
    CommandLine line(static_cast<int>(args.size()), &args.begin()[0]);

    // Supported command line switches:
    //      -f = EXO filename
    //      -t = target number
    //
    //  ...Clear "instruction" variables
    Settings.Auto = args.size() > 1;
    if (!Settings.Auto)
        return;

    if (line.search("-f"))
        {
        Settings.ImageFile = line.next("intf.bit");
        Settings.Auto = true;
        }

    if (line.search("-t"))
        Settings.Target = StringToInt(line.next("0"));

    Timer.Enabled(Settings.Auto);
}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Open() -- Open target
//---------------------------------------------------------------------------

void ApplicationIo::Open()
{
    try
        {
        Baseboard.BusMasterSize(1<<20);
        Baseboard.Target(Settings.Target);
        Baseboard.Open();
        UI->Log("Target driver opened ok");
        DisplayLogicVersion();
        FOpen = true;
        }
    catch(...)
        {
        UI->Log("Could not open target ");
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*Event*/)
{
    if (Settings.Auto)
        {
        Open();
        if (FOpen)
            ConfigureFpga(Settings.ImageFile);
        }

    Timer.Enabled(false);
	Timer.OnElapsed.ClearEvent();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Close() -- Close target
//---------------------------------------------------------------------------

void ApplicationIo::Close()
{
    Baseboard.Close();
    FOpen = false;
}

//------------------------------------------------------------------------------
//  ApplicationIo::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  ApplicationIo::DisplayLogicVersion()
{
    char pcb = 'A' + static_cast<char>(Baseboard.UserLogicVersion().FpgaLogicRevision());

    std::stringstream msg;
    msg << std::hex << "Logic Version: " << Baseboard.UserLogicVersion().FpgaLogicVersion()
        << ", Variant: " << Baseboard.UserLogicVersion().FpgaLogicVariant()
        << ", Hardware Version: " << pcb
        << ", Type: " << Baseboard.UserLogicVersion().FpgaLogicType();
    UI->OnRevision(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::ConfigureFpga() -- Load logic from file
//---------------------------------------------------------------------------

void  ApplicationIo::ConfigureFpga(const std::string & exo_file)
{
    return Baseboard.Logic().ConfigureFpga(exo_file);
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- return # of boards found on bus
//---------------------------------------------------------------------------

size_t ApplicationIo::BoardCount()
{
    return Baseboard.BoardCount();
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardName() --  Return board name string
//---------------------------------------------------------------------------

std::string ApplicationIo::BoardName()
{
    return PmcDriver;
}

//---------------------------------------------------------------------------
//  ApplicationIo::LogicVersion() --  Return reference to EEPROM version object
//---------------------------------------------------------------------------

Innovative::FpgaLogicVersionInfoIntf & ApplicationIo::LogicVersion()
{
    return Baseboard.UserLogicVersion();
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleParseProgress() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleParseProgress(Innovative::ProcessProgressEvent & event)
{
    UI->OnProgress(event.Percent);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleParseComplete() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleParseComplete(Innovative::ProcessCompletionEvent & /*event*/)
{
    UI->OnParseComplete();
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadProgress() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadProgress(ProcessProgressEvent & event)
{
    UI->OnProgress(event.Percent);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadError() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadError(ProcessStatusEvent & event)
{
    UI->OnLoadError(event.Message);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadComplete() --
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadComplete(ProcessCompletionEvent & /*event*/)
{
    if (FOpen && Settings.Auto)
        Close();

    UI->OnLoadComplete();
}

//------------------------------------------------------------------------------
//  ApplicationIo::Sleep() --
//------------------------------------------------------------------------------

void ApplicationIo::Sleep(unsigned int ms)
{  
    Innovative::Sleep(ms);  
}


//==============================================================================
//  CLASS ApplicationSettings - Remember Settings between runs
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class ApplicationSettings --  load from INI file
//------------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings()
{
	Application App;
	std::string file = App.Path() + std::string("Settings.ini");
	Ini = new IniFile(file);

    Target  = Ini->Read("Target", 0);
    ImageFile = Ini->Read("ImageFile", std::string(""));
}

//------------------------------------------------------------------------------
//  destructor for class ApplicationSettings --  save to INI file
//------------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    Ini->Write("Target",  Target);
    Ini->Write("ImageFile",  ImageFile);

	delete Ini;
	Ini = 0;
}

