// ApplicationIo.cpp
//
// Board-specific data flow and hardware I/O
// Copyright 2006 Innovative Integration
//---------------------------------------------------------------------------

#include <Malibu_Mb.h>
#include <Analysis_Mb.h>

#include <Exception_Mb.h>
#include "ApplicationIo.h"
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <sstream>
#include <limits>
#include <cmath>
#include <numeric>
#include <Random_Mb.h>
#include <Magic_Mb.h>
#include <FileSupport_Mb.h>
#include <DeviceEnum_Mb.h>
#include <IppSupport_Mb.h>
#include <Buffer_Mb.h>
#include <BufferHeader_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <IppMemoryUtils_Mb.h>

using namespace Innovative;
using namespace std;

//===========================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    : FiclIo(ui), UI(ui), Pip(&Stream), Cursor(0), Opened(false),
      StreamConnected(false), Time(5), BytesPerBlock(6), FLogicLoaded(false),
      FLogicError(false)
{
    TraceVerbosity(Trace::vNormal);

    OnLog.SetEvent(this, &ApplicationIo::HandleOnLog);

    // Use IPP performance memory functions.
    Init::UsePerformanceMemoryFunctions();

    //
    //  Set up Loggers and graphs
    std::stringstream ss;
    ss << Settings.Root() << "Data.bin";

    Logger.FileName(ss.str());
    Graph.BinFile(Logger.FileName());

    Timer.Interval(1000);

    std::stringstream msg;
    msg << "Be sure to read the help file for info on this program";
    msg << " located in the root of the this example folder.  ";

    Log(msg.str());

    if (Settings.Help)
        {
        UI->DisplayHelp();
        Settings.Help = false;
        }
}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::FiclInit() -- Set up FICL system
//---------------------------------------------------------------------------

void  ApplicationIo::FiclInit()
{
    //  Link Module to Ficl System
    Module.FiclConnectTo(FiclSys());

    //  Start Ficl System
    FiclIo_Init();
}


//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

unsigned int ApplicationIo::BoardCount()
{
    return static_cast<int>(Module().BoardCount());
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardNames() -- Get Device Names
//---------------------------------------------------------------------------

vector<string> ApplicationIo::BoardNames()
{
    return Module().BoardNames();
}

//---------------------------------------------------------------------------
//  ApplicationIo::SetTargetIndex() -- Set Target Index
//---------------------------------------------------------------------------

unsigned int ApplicationIo::SetTargetIndex()
{
    vector<string> BNames = BoardNames();
    for (size_t i=0; i<BNames.size(); i++)
        {
        string Name = BNames[i];
        size_t found = Name.find("X3 XMCe");
        if(found!=string::npos)
            {
            return i;
            }
        }

    return 0;
}

//---------------------------------------------------------------------------
//  ApplicationIo::PrintDevices() -- Print Device Names and Target #
//---------------------------------------------------------------------------

string ApplicationIo::PrintDevices()
{
    return Module().PrintAllDevices();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Open() -- Open Hardware & set up callbacks
//---------------------------------------------------------------------------

void  ApplicationIo::Open()
{
    static bool once = false;
    if (!once && FileExists(Settings.StartupScript))
        {
        Execute(std::string("load ") + Settings.StartupScript);
        once = true;
        }

    UI->GetSettings();

    Module().Clock().SetPhaseDetectorFrequency(Settings.PhaseDetectorFrequency);

    //
    //  Configure Module Event Handlers
    Module().Logic().OnFpgaFileReadProgress.SetEvent(this, &ApplicationIo::HandleProgress);
    Module().Logic().OnFpgaFileReadComplete.SetEvent(this, &ApplicationIo::HandleParseComplete);
    Module().Logic().OnFpgaParseProgress.SetEvent(this, &ApplicationIo::HandleProgress);
    Module().Logic().OnFpgaParseComplete.SetEvent(this, &ApplicationIo::HandleParseComplete);
    Module().Logic().OnFpgaParseMessage.SetEvent(this, &ApplicationIo::HandleLoadError);
    Module().Logic().OnFpgaLoadProgress.SetEvent(this, &ApplicationIo::HandleProgress);
    Module().Logic().OnFpgaLoadComplete.SetEvent(this, &ApplicationIo::HandleLoadComplete);
    Module().Logic().OnFpgaLoadMessage.SetEvent(this, &ApplicationIo::HandleLoadError);

    Module().OnBeforeStreamStart.SetEvent(this, &ApplicationIo::HandleBeforeStreamStart);
    Module().OnBeforeStreamStart.Synchronize();
    Module().OnAfterStreamStart.SetEvent(this, &ApplicationIo::HandleAfterStreamStart);
    Module().OnAfterStreamStart.Synchronize();
    Module().OnAfterStreamStop.SetEvent(this, &ApplicationIo::HandleAfterStreamStop);
    Module().OnAfterStreamStop.Synchronize();

    //
    //  Alerts
    Module().Alerts().OnTimeStampRolloverAlert.SetEvent(this, &ApplicationIo::HandleTimestampRolloverAlert);
    Module().Alerts().OnSoftwareAlert.SetEvent(this, &ApplicationIo::HandleSoftwareAlert);
    Module().Alerts().OnWarningTemperature.SetEvent(this, &ApplicationIo::HandleWarningTempAlert);
    Module().Alerts().OnPllLost.SetEvent(this, &ApplicationIo::HandlePllLostAlert);
    Module().Alerts().OnInputFifoOverrun.SetEvent(this, &ApplicationIo::HandleInputFifoOverrunAlert);
    Module().Alerts().OnInputTrigger.SetEvent(this, &ApplicationIo::HandleInputTriggerAlert);
    Module().Alerts().OnInputOverrange.SetEvent(this, &ApplicationIo::HandleInputFifoOverrangeAlert);
    Module().Alerts().OnInputStopTrigger.SetEvent(this, &ApplicationIo::HandleInputStopTriggerAlert);
    Module().Alerts().OnInputPretriggerError.SetEvent(this, &ApplicationIo::HandleInputPretriggerErrorAlert);

#if 1
    //
    //  Configure Stream Event Handlers
    Stream.OnDataAvailable.SetEvent(this, &ApplicationIo::HandleDataAvailable);
    Stream.DirectDataMode(false);
#else
    //
    //  Configure Stream Event Handlers
    // This particular handler runs in the original background thread
    // context, to eliminate thunking and maximize performance
    // Additionally, direct data mode is used to mitigate CPU time spent
    // administring the received data
    Stream.OnDirectDataAvailable.SetEvent(this, &ApplicationIo::HandleDirectDataAvailable);
    Stream.DirectDataMode(true);

    // This event allows for parsing and processing of packet data even though were
    // using DirectData mode (above).  Uncomment if needed
    Pip.OnImageAvailable.SetEvent(this, &ApplicationIo::HandlePipProcessing);
#endif

    Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);
    Timer.OnElapsed.Thunk();

    // Insure BM size is a multiple of four MB
    const int Meg = 1024 * 1024;
    const int BmSize = max(Settings.BusmasterSize/4, 1) * 4;
    Module().IncomingBusMasterSize(BmSize * Meg);
    Module().OutgoingBusMasterSize(1 * Meg);
    Module().Target(Settings.Target);

    // Outbound requirements are minimal in this example
    Module().OutgoingBusMasterSize(1 * Meg);
    Module().Target(Settings.Target);

    // Open Device
    try
        {
        Module().Open();
        }

    catch (Innovative::MalibuException & e)
        {
        UI->Status("Module Device Open Failure!");
        UI->Log("Open Failure:");
        UI->Log(e.what());
        Opened = false;
        return;
        }

    catch(...)
        {
        UI->Status("Module open failure...");
        Opened = false;
        return;
        }


    Module().Reset();
    UI->Status("Module device opened successfully...");
    Opened = true;
    //
    //  Connect Stream
    Stream.ConnectTo(&(Module.Ref()));
    StreamConnected = true;
    UI->Status("Stream Connected...");

    //  Perform Board Specific Initialization
    Module.AfterOpen();

    DisplayLogicVersion();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Close() -- Close Hardware & set up callbacks
//---------------------------------------------------------------------------

void  ApplicationIo::Close()
{
    if (!Opened)
        return;

    Stream.Disconnect();
    StreamConnected = false;

    Module().Close();
    Opened = false;

    UI->Status("Stream Disconnected...");
}

//---------------------------------------------------------------------------
//  ApplicationIo::CalculateExternalFrequency() --  Ext clock for specified rate
//---------------------------------------------------------------------------

void ApplicationIo::CalculateExternalFrequency()
{
    if (Stopped)
        {
        UI->GetSettings();

        Module().Clock().Frequency(SampleRate());
        float result = static_cast<float>(SampleRate() * Module().Input().Info().ClockFactor() / 1.e3);

        std::stringstream ss;
        ss << "calculated external clock rate is " << result;
        UI->Log(ss.str());
       }
}

//---------------------------------------------------------------------------
//  ApplicationIo::StartStreaming() --  Initiate data flow
//---------------------------------------------------------------------------

void  ApplicationIo::StartStreaming()
{
    if (!StreamConnected)
        {
        Log("Stream not connected! -- Open the boards");
        return;
        }

    //
    //  Set up Parameters for Data Streaming
    //  ...First have UI get settings into our settings store
    UI->GetSettings();

    //Set the pfd error
    //Module().Clock().PfdError(50.0f);

    if (SampleRate() > Module().Input().Info().MaxRate())
        {
        Log("Sample rate too high: Max Rate =", Module().Input().Info().MaxRate());
        StopStreaming();
        UI->AfterStreamAutoStop();
        return;
        }
    if (SampleRate() < Module().Input().Info().MinRate())
        {
        Log("Sample rate too low: Min Rate =", Module().Input().Info().MinRate());
        StopStreaming();
        UI->AfterStreamAutoStop();
        return;
        }

    if (Settings.Framed)
        {
        // Avoid residual data in pipeline between frames
        unsigned int PretriggerSize = Settings.PretriggerEnable ? Settings.PretriggerCount : 0u;
        unsigned int TriggerSize = Settings.FrameSize+PretriggerSize;
        if (Settings.PacketSize != TriggerSize)
            Log("Warning: Packet/frame size mismatch will irregulate data flow.");
        }

    FBlockCount = 0;
    FBlockRate = 0;
    FTriggered = -1;

    //  Set external trigger source (front or P16)
    Module().Input().FrontPanelTrigger(Settings.FrontPanelTrigger);

    // Disable triggering initially
    Module().Input().ExternalTrigger(false);
    Module.SetInputSoftwareTrigger(false);

    // Set sync source
    Module().Master(Settings.SyncMaster ? true : false);

    //
    //  Channel Enables
    Module().Input().Info().Channels().DisableAll();
    int userChannelCount = 0;
    int range(0);
    bool multirange = false;
    for (unsigned int i = 0; i < Channels(); ++i)
    {
        if (Settings.ActiveChannels[i])
        {
            if (userChannelCount)
            {
                if (range != Settings.Range[i])
                    multirange = true;
            }

            range = Settings.Range[i];

            ++userChannelCount;
            Module().Input().Info().Channels().Enabled(i, true);
            Module().Input().Range(i, static_cast<X3RangedIoDevice::IIRange>(range));

            // Adjust graph range
            Graph.InputSpan(static_cast<float>(Module().Input().Info().Span()[Settings.Range[i]].Delta()));
        }
    }

    if (multirange)
        Log("Warning: Multiple analog gain ranges in use.  Binview scaling will be incorrect for some channels.");

    int ActiveChannels = Module().Input().Info().Channels().ActiveChannels();
    int ActiveDevices = Module().Input().Info().Channels().ActiveDevices();
    int GangSize = Module().Input().Info().GangSize();
    if (!ActiveChannels)
        {
        Log("Error: Must enable at least one channel");
        UI->AfterStreamAutoStop();
        return;
        }

    // Warn of device enable granularity violations
    if (userChannelCount != ActiveDevices*GangSize)
        {
        std::stringstream msg;
        msg << "NOTE:  Streaming " << ActiveDevices*GangSize << " instead of "
            << userChannelCount << " channels.";
        Log(msg.str());
        }

    // Packets scaled in units of events (samples per each enabled channel)
    int SamplesPerWord = 4 / Module().Input().Info().Channels().BytesPerSample();
    int packetsize = Settings.PacketSize*ActiveChannels/SamplesPerWord + 2;
    int FBusmasterSize = 1 << (Settings.BusmasterSize + 22);
    if (static_cast<int>(packetsize*sizeof(int)) > (FBusmasterSize/2))
        {
        Log("Packet size too large for current busmaster region");
        StopStreaming();
        UI->AfterStreamAutoStop();
        return;
        }

    Module().Input().PacketSize(packetsize);

    //
    //  Start Loggers on active channels
    if (Settings.PlotEnable)
        Graph.Quit();

    if (Settings.RTPlotEnable)
        {
        Plot.Stop();

        Plot.Bdd.Channels(Module().Input().Info().Channels().ActiveChannels());
        Plot.Bdd.SignificantBits(Module().Input().Info().Channels().Bits());
        Plot.Bdd.Samples(Settings.RTPlotSamples* Module().Input().Info().Channels().BytesPerEvent());
        Plot.Bdd.SampleRate(static_cast<float>(SampleRate()/1.0e3));
        Plot.UpdateRate(Settings.RTPlotFrequency);

        Plot.Start();
        }

    TriggerTime = 0;
    TriggerLogger.FileName(Settings.Root()+"TriggerLog.bin");
    TriggerLogger.Start();

    StopTriggerTime = 0;
    StopTriggerLogger.FileName(Settings.Root()+"StopTriggerLog.bin");
    StopTriggerLogger.Start();

    if (Settings.LoggerEnable || Settings.PlotEnable)
        Logger.Start();

    BlocksToLog = Settings.SamplesToLog/Settings.PacketSize
        + ((Settings.SamplesToLog%Settings.PacketSize) ? 1 : 0);

    Stopped = false;

    Module().Dio().DioPortConfig(Settings.DioConfig);
    Module().FrontPanel().FrontPanelPortConfig(Settings.FrontPanelConfig);

    // Set test mode
    Module().TestCounterEnable(Settings.TestCounterEnable);


    if (Settings.PretriggerEnable)
        {
        unsigned int PTrigCountWords = Settings.PretriggerCount*ActiveChannels/SamplesPerWord;
        //
        //  Check pretrigger validity
        if (PTrigCountWords > Module().Input().MaxPretriggerCountWords())
            {
            unsigned int MaxPT = Module().Input().MaxPretriggerCountWords();

            Log("Pretrigger Count too large: Max =", MaxPT*SamplesPerWord/ActiveChannels);
            StopStreaming();
            UI->AfterStreamAutoStop();
            return;
            }
        Module().Input().PretriggerCount(PTrigCountWords);
        }
    else
        Module().Input().PretriggerCount(0);
    Module().Input().PretriggerMode(Settings.PretriggerEnable);
    Module().Input().TriggerDelay(Settings.TriggerDelay);

    // Initialize PIP, in case DirectData mode is in effect
    Pip.Clear();

    // Route clock to active analog devices
    // Set reference based on clock source to obtain correct FrequencyActual
    double reference = ReferenceRate();
    switch (Settings.SampleClockSource)
        {
        case 0: // J1
            Module().Clock().OutputClock(ModuleIo::Timebase::oExternal);
            Module().ExternalClockSource(ModuleIo::ModuleType::esPanel);
            break;
        case 1: // P16
            Module().Clock().OutputClock(ModuleIo::Timebase::oExternal);
            Module().ExternalClockSource(ModuleIo::ModuleType::esPxi);
            break;
        case 2: // PLL
            {
            Module().Clock().OutputClock(ModuleIo::Timebase::oVco);
            if (Settings.ReferenceSource == 0)
                reference = Module().Input().Info().ReferenceClock();
            }
            break;
        }

    // Apply timebase correction factor, if available
    double correction = Settings.PllCorrection;
    if (correction != correction)
        correction = 1.0;	// NaN, so fix it
    Module().Clock().ReferenceCalibrationFactor(correction);


    // Route reference.
    ModuleIo::ModuleType::IIReferenceSource ref[] = { ModuleIo::ModuleType::rsCrystal, ModuleIo::ModuleType::rsPxi };
    Module().ReferenceSource(ref[Settings.ReferenceSource]);
    Module().Clock().Reference(reference);
    Module().Clock().Frequency(SampleRate());

    // Set Decimation Factor, using AD9511 calculated settings
    unsigned int DecimationFactor = Module().Clock().Decimation();
    //Module().Output().InterpolationFactor(DecimationFactor);
    Module().Input().Decimation(DecimationFactor);
    if (Module().Clock().Decimation() != 1)
        {
        Log("Decimation applied to reach the requested frequency.");
        }

    // Frame count in units of packet elements
    if (Settings.Framed)
        Module().Input().Framed(Settings.FrameSize);
    else
        Module().Input().Unframed();

    enum IUsesX3Alerts::AlertType Alert[] = {
        IUsesX3Alerts::alertTimeStampRollover, IUsesX3Alerts::alertSoftware,
        IUsesX3Alerts::alertWarningTemperature,
        IUsesX3Alerts::alertPllLost, IUsesX3Alerts::alertInputFifoOverrun,
        IUsesX3Alerts::alertInputTrigger, IUsesX3Alerts::alertInputOverrange,
        IUsesX3Alerts::alertInputPretriggerError, IUsesX3Alerts::alertInputStopTrigger };

    for (unsigned int i = 0; i < Settings.AlertEnable.size(); ++i)
        Module().Alerts().AlertEnable(Alert[i], Settings.AlertEnable[i] ? true : false);

    //
    //  Start Streaming
    Stream.Start();
    Log("Stream Mode started");
    UI->Status("Stream Mode started");

    ActualSampleRate = static_cast<float>(Module().Clock().FrequencyActual());

    std::stringstream msg;
    msg.precision(6);
        msg << "Actual sampling rate: " << ActualSampleRate/(1.e3 * Module().Clock().Decimation()) << " KHz";
    Log(msg.str());

    FTicks = 0;
    Timer.Enabled(true);
}

//---------------------------------------------------------------------------
//  ApplicationIo::StopStreaming() --  Terminate data flow
//---------------------------------------------------------------------------

void  ApplicationIo::StopStreaming()
{
    if (!StreamConnected)
        {
        Log("Stream not connected! -- Open the boards");
        return;
        }

    //
    //  Stop Streaming
    Stream.Stop();
    Stopped = true;
    Timer.Enabled(false);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Configuration Event Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//------------------------------------------------------------------------------
//  ApplicationIo::HandleParseComplete() --  Incremental logic parse event
//------------------------------------------------------------------------------

void  ApplicationIo::HandleParseComplete(ProcessCompletionEvent & /*event*/)
{
    Log("Parse completed ok");
}

//--------------------------------------------------------------------------
//  ApplicationIo::HandleProgress() --  Incremental logic load event
//--------------------------------------------------------------------------

void  ApplicationIo::HandleProgress(ProcessProgressEvent & event)
{
    UI->UpdateLogicLoadProgress(event.Percent);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadComplete() --  Logic load completion event
//------------------------------------------------------------------------------

void  ApplicationIo::HandleLoadComplete(ProcessCompletionEvent & /*event*/)
{
    Log("Load completed ok");

    DisplayLogicVersion();
    FLogicLoaded = true;
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadError() --  Logic load failure event
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadError(ProcessStatusEvent & event)
{
    Log(event.Message);
    FLogicError = true;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Data Flow Event Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  ApplicationIo::HandleDataAvailable() --  Handle received packet
//---------------------------------------------------------------------------

void  ApplicationIo::HandleDataAvailable(PacketStreamDataEvent & Event)
{
    if (Stopped)
        return;

    Buffer Packet;
    //
    //  Extract the packet from the Incoming Queue...
    Event.Sender->Recv(Packet);
    IntegerDG Packet_DG(Packet);

    if (Settings.RTPlotEnable)
        Plot.Update(Packet);

    //
    //  Process the data packet
    PacketBufferHeader PktBufferHdr(Packet);

    // If enabled, log the data stream
    if (Settings.LoggerEnable || Settings.PlotEnable)
        if (FBlockCount < BlocksToLog)
            Logger.LogWithHeader(Packet);

    TallyBlock(Packet_DG.size()*sizeof(int));
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleDirectDataAvailable() --  Log data direct from BM
//---------------------------------------------------------------------------

void  ApplicationIo::HandleDirectDataAvailable(PacketStreamDirectDataEvent & Event)
{
    if (Stopped)
        return;

    // Calculate transfer rate in MB/s
    size_t Ints = Event.Slab.SizeInInts();
    const int * Address = reinterpret_cast<const int*>(Event.Slab.Address());

    // If enabled, log the data stream
    if (Settings.LoggerEnable || Settings.PlotEnable)
        if (FBlockCount < BlocksToLog)
            Logger.Log(Address, static_cast<unsigned int>(Ints));

    //
    //  Run the data thru a packet parser
    //  Allows processing of data on packet boundaries if needed
    Pip.Append(Event.Slab.Image);
    Pip.Parse();

    //  We have processed all the events in the slab..grant credit for
    //     that amount now. Any residual will be acked next time
    Event.Sender->DirectDataReceiveAck(Pip.ProcessedInInts());
    Pip.ResetProcessed();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TallyBlock() --  Finish processing of received packet
//---------------------------------------------------------------------------

void  ApplicationIo::TallyBlock(size_t bytes)
{
    double Period = Time.Differential();
    double AvgBytes = BytesPerBlock.Process(static_cast<double>(bytes));
    if (Period)
        FBlockRate = AvgBytes / (Period*1.0e3);

    //
    //  Tally the blocks received
    ++FBlockCount;

    //
    //  Stop streaming when both Channels have passed their limit
    if (Settings.AutoStop && IsDataLoggingCompleted() && !Stopped)
        {
        // Stop counter and display it
        double elapsed = Clock.Stop();

        StopStreaming();
        Log("Stream Mode Stopped automatically");
        Log(std::string("Elasped (S): ") + FloatToString(elapsed));
        }

    // Auto-analyze and retrigger in framed mode
    if (!Settings.Framed)
        return;

    if ((Settings.ExternalTrigger == false) && Settings.AutoTrigger)
        {
        __int64 samples = FBlockCount * Settings.PacketSize;
        int triggers = static_cast<int>(samples/Settings.FrameSize);

        if (triggers != FTriggered)
            SoftwareTrigger();
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandlePipProcessing() --  Process Parsed Packet
//---------------------------------------------------------------------------

void  ApplicationIo::HandlePipProcessing(PacketImageParserImageAvailable & Event)
{
    // PID of data to be processed
    const unsigned int Pid = 0;

   //  Check if packet PID matches our PID
   unsigned int pid = Event.Image.PeripheralId();
    if (pid == Pid)
        {
        Innovative::Buffer buffer;

        //  Fill buffer up to our size limit, or the packet size, whichever is smaller
        Event.Image.CopyDataTo(buffer);

        // Insert your packet processing here...
        if (Settings.RTPlotEnable)
            Plot.Update(buffer);

        IntegerDG Packet_DG(buffer);
        TallyBlock(Packet_DG.SizeInBytes());
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleBeforeStreamStart() --  Pre-streaming init event
//---------------------------------------------------------------------------

void ApplicationIo::HandleBeforeStreamStart(OpenWire::NotifyEvent & /*event*/)
{
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleAfterStreamStart() --  Post streaming init event
//---------------------------------------------------------------------------

void ApplicationIo::HandleAfterStreamStart(OpenWire::NotifyEvent & /*event*/)
{
    Log(std::string("Analog I/O started"));
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleAfterStreamStop() --  Post stream termination event
//---------------------------------------------------------------------------

void ApplicationIo::HandleAfterStreamStop(OpenWire::NotifyEvent & /*event*/)
{
    Module().Trigger(ModuleIo::ModuleType::tInput, false);
    Module().Input().ExternalTrigger(false);

    TriggerLogger.Stop();
    StopTriggerLogger.Stop();
    //
    //  Stop Loggers on active Channels
    if (Settings.LoggerEnable || Settings.PlotEnable)
        {
           Logger.Stop();
           InitBddFile();

           if (Logger.Logged() && Settings.PlotEnable)
               Graph.Plot();
        }

    UI->AfterStreamAutoStop();
    Log(std::string("Analog I/O Stopped"));
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*event*/)
{
    // Wiggle DIO bits to illustration functionality
    int DigIn = DioData();
    int FrontPanelIn = FrontPanelData();

    // Display status
    UI->PeriodicStatus();

    FrontPanelData(~FrontPanelIn);
    DioData(~DigIn);

    // Initial trigger state machine below
    if (IsTriggered())
        return;

    // Give PLL up to 2 secs to lock prior to triggering
    if (++FTicks < 2)
        return;

    if (Settings.ExternalTrigger == false)
        SoftwareTrigger();
    else
        {
        Module().Input().ExternalTrigger(true);
        ++FTriggered;
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const std::string & msg, double argument)
{
    std::stringstream text;
    text << msg << argument;
    Log(text.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const std::string & msg, unsigned int argument)
{
    std::stringstream text;
    text << msg << IntToString(argument, 16);
    Log(text.str());
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Alert Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimestampRolloverAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleTimestampRolloverAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Time stamp rollover " << event.Argument
        << " after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleSoftwareAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleSoftwareAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Software alert 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleWarningTempAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleWarningTempAlert(Innovative::AlertSignalEvent & event)
{

    std::stringstream msg;
    msg << "Temp warning alert 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);

    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandlePllLostAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandlePllLostAlert(Innovative::AlertSignalEvent & event)
{
    // Ignore during PLL stabilization
    if (FTicks < 2)
        return;

    std::stringstream msg;
    msg << "PLL lost sync after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleInputFifoOverrunAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleInputFifoOverrunAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Input FIFO overrun after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleInputTriggerAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleInputTriggerAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Input triggered after " << Elapsed(event.TimeStamp);
    Log(msg.str());
 /*
    size_t delta = event.TimeStamp-TriggerTime;
    // Log time at which point trigger was received
    TriggerLogger.Log(reinterpret_cast<const int *>(&delta), 1);
    TriggerTime = event.TimeStamp;
*/
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleInputStopTriggerAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleInputStopTriggerAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Input stop trigger after " << Elapsed(event.TimeStamp);
    Log(msg.str());
/*
    size_t delta = event.TimeStamp-StopTriggerTime;
    // Log time at which point trigger was received
    StopTriggerLogger.Log(reinterpret_cast<const int *>(&delta), 1);
    StopTriggerTime = event.TimeStamp;
*/
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleInputPretriggerErrorAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleInputPretriggerErrorAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Pretrigger Error: Only 0x" << std::hex << event.Argument
        << " words captured at " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleInputFifoOverrangeAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleInputFifoOverrangeAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Input overranged after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOnLog() --
//---------------------------------------------------------------------------

void ApplicationIo::HandleOnLog(Innovative::ProcessStatusEvent & Event)
{
    UI->Log(Event.Message);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Support Functions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  ApplicationIo::IsDataLoggingCompleted() --check if done logging
//---------------------------------------------------------------------------

bool  ApplicationIo::IsDataLoggingCompleted()
{
    if (!Settings.LoggerEnable)
        return true;

    return FBlockCount >= BlocksToLog;
}

//---------------------------------------------------------------------------
//  ApplicationIo::LoadLogic() -- Initiate Logic Load Process
//---------------------------------------------------------------------------

void  ApplicationIo::LoadLogic()
{
    if (!Opened)
        {
        Log("No module on specified target");
        return;
        }

    Log("-----------------------------------------------");
    Log("  Parsing Module logic file");

    UI->GetSettings();
    Module().Logic().ConfigureFpga(Settings.ExoFile);
}

//---------------------------------------------------------------------------
//  ApplicationIo::SoftwareAlert() --  Issue specified software alert
//---------------------------------------------------------------------------

void ApplicationIo::SoftwareAlert(unsigned int value)
{
    Log(std::string("Posting SW alert..."));
    Module().Alerts().SoftwareAlert(value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

float ApplicationIo::Temperature()
{
    return Module().Thermal().LogicTemperature();
}

//---------------------------------------------------------------------------
//  ApplicationIo::DioData() --  Set DIO bits state
//---------------------------------------------------------------------------

void ApplicationIo::DioData(unsigned int value)
{
    return Module().Dio().DioPortData().Value(value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DioData() --  Query DIO bits state
//---------------------------------------------------------------------------

unsigned int ApplicationIo::DioData()
{
    return Module().Dio().DioPortData().Value();
}

//---------------------------------------------------------------------------
//  ApplicationIo::FrontPanelData() --  Set front panel bits state
//---------------------------------------------------------------------------

void ApplicationIo::FrontPanelData(unsigned int value)
{
    Module().FrontPanel().FrontPanelPortData().Value(value);
}
//---------------------------------------------------------------------------
//  ApplicationIo::FrontPanelData() -- Query front panel bits state
//---------------------------------------------------------------------------

unsigned int ApplicationIo::FrontPanelData()
{
    return Module().FrontPanel().FrontPanelPortData().Value() & 0xfff;
}

//---------------------------------------------------------------------------
//  ApplicationIo::Pll_Locked() -- Query front panel bits state
//---------------------------------------------------------------------------

bool ApplicationIo::Pll_Locked()
{
    return Module().Clock().Locked();
}

//---------------------------------------------------------------------------
//  ApplicationIo::SoftwareTrigger() --  Issue single software trigger pulse
//------------------------------------------------------------------------------

void  ApplicationIo::SoftwareTrigger()
{
    Module().Trigger(ModuleIo::ModuleType::tInput, false);
    Module().Trigger(ModuleIo::ModuleType::tInput, true);
    ++FTriggered;
}

//---------------------------------------------------------------------------
//  ApplicationIo::PretriggerSupported() -- Is Pretriggering supported?
//---------------------------------------------------------------------------

bool ApplicationIo::PretriggerSupported()
{
    return Module().Input().Info().HasPretriggering();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TriggerDelaySupported() -- Is Pretriggering supported?
//---------------------------------------------------------------------------

bool ApplicationIo::TriggerDelaySupported()
{
    return Module().Input().Info().HasTriggerDelay();
}

//------------------------------------------------------------------------------
//  ApplicationIo::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  ApplicationIo::DisplayLogicVersion()
{
    std::stringstream msg;
    msg << std::hex << "Version: " << Module().UserLogicVersion().FpgaLogicVersion()
        << ", Variant: " << Module().UserLogicVersion().FpgaLogicVariant()
        << ", Revision: " << Module().UserLogicVersion().FpgaLogicRevision()
        << ", Type: " << Module().UserLogicVersion().FpgaLogicType();
    UI->LogicStatus(msg.str());
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleName() --  Return name of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleName() const
{
    return ModuleNameStr();
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleRevision() --  Return rev of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleRevision()
{
    std::stringstream msg;
    char c = 'A' + static_cast<char>(Module().UserLogicVersion().FpgaLogicRevision());
    msg << c;
    return msg.str();
}

//------------------------------------------------------------------------------
//  ApplicationIo::Elapsed() --  Display timestamp as elapsed MCLKs
//------------------------------------------------------------------------------

std::string ApplicationIo::Elapsed(size_t timestamp)
{
    stringstream msg;
    msg << timestamp << " master clocks";
    return msg.str();
}

//---------------------------------------------------------------------------
//  ApplicationIo::InitBddFile() --  Set up BDD File for an output file
//---------------------------------------------------------------------------

void ApplicationIo::InitBddFile()
{
    BinView & Bv(Graph);
    UI->GetSettings();

    std::stringstream msg;
    msg << "X3Snap";
    Bv.System().ServerSlotName(msg.str());

    // Update only if no BDD presently
    PathSpec spec(Bv.BinFile());
    spec.Ext(".bdd");

    if (FileExists(spec.Full()) && !Settings.OverwriteBdd)
        return;

    // Make a new BDD
    const int bits = Module().Input().Info().Channels().Bits();

    const int limit = 1 << (bits-1);

    Bv.Time().LowerLimit(-limit);
    Bv.Time().UpperLimit(limit);
    Bv.Time().Break(100);
    Bv.Time().SeamSize((Module().Input().PacketSize()-2)*sizeof(int));
    Bv.Time().AnnotateSeams(true);
    Bv.Fft().LowerLimit(-150);
    Bv.Fft().ScaleYType(BinViewOptions::FftOptions::sLog);
    Bv.Fft().ScaleXType(BinViewOptions::FftOptions::sLog);
    Bv.Fft().CriteriaThreshold(75);
    BinViewOptions::FftOptions::IIPoints pts(static_cast<BinViewOptions::FftOptions::IIPoints>(
        TrailingZeroCount(Settings.PacketSize)-7));
    Bv.Fft().Points(pts);
    Bv.Fft().Window(BinViewOptions::FftOptions::wBlackmann);
    Bv.Fft().Average(1);
    Bv.Text().DataFormat("%11.3f");
    Bv.System().UpdateRate(BinViewOptions::SystemOptions::ms1000);
    Bv.System().ServerSlotName("BinView");
    Bv.Leap(Settings.PacketSize);
    Bv.SignificantBits(bits);
    Bv.Polarity(BinView::pSigned);
    Bv.DataSpan(300);
    Bv.DataType(BinView::tInt);
    Bv.Units("mV");
    Bv.Samples(1000);
    Bv.Channels(Module().Input().Info().Channels().ActiveChannels());
    Bv.Devices(1);
    Bv.System().Source(BinViewOptions::SystemOptions::sPacketFile);
    Bv.NullDataValue(.12345e-19f);
    Bv.SampleRate(static_cast<float>(ActualSampleRate/1000.));

    // Use scalefactor from first enabled channel
    for (unsigned int i = 0; i < Channels(); ++i)
        {
        if (Settings.ActiveChannels[i])
            {
            int range = Settings.Range[0];
            float span = static_cast<float>(Module().Input().Info().Span()[range].Delta());
            Bv.InputSpan(span);
            break;
            }
        }

    Bv.ScalingEnabled(true);

    Bv.Save();
}

//------------------------------------------------------------------------
//  ApplicationIo::TestZbt() -- Test ZBT RAM
//------------------------------------------------------------------------

void ApplicationIo::TestZbt()
{
    size_t percent[2] = {(size_t)0, (size_t)-1};
    std::vector<int> Buffer(0x100000 >> 1);

    // Fill ZBT RAM from buffer filled with random, 32-bit values
    double One = std::pow(2.0, 32);
    size_t count = Buffer.size();
    Innovative::Random x;
    for (size_t i = 0; i < count; ++i)
        {
        Buffer[i] = static_cast<int>(x.uniform() * One);
        Module().Ram[static_cast<int>(i)] = Buffer[i];

        // Display progress
        percent[0] = (i*100)/count;
        if (percent[0] == percent[1])
            continue;

        std::stringstream msg;
        msg << "Zbt ram pattern build %" << percent[0] << " complete...";
        UI->Status(std::string(msg.str()));
        percent[1] = percent[0];
        }

    // Compare ZBT RAM contents to original buffer, reporting errors
    int errors = 0;
    for (unsigned int i = 0; i < count; ++i)
        {
        unsigned short is = Module().Ram[i];
        unsigned short shouldbe = Buffer[i];

        int differ = (is != shouldbe) ? 1 : 0;
        if (differ && (errors < 10))
            {
            std::stringstream msg;
            msg << std::hex << "Addr " << i << " is " << is
                << ", should be " << shouldbe
                << " (XOR is " << (shouldbe ^ is) << ")";
            Log(std::string(msg.str()));
            }

        errors += differ;

        // Display progress
        percent[0] = (i*100)/count;
        if (percent[0] == percent[1])
            continue;

        std::stringstream msg;
        msg << "Zbt ram test %" << percent[0] << " complete...";
        UI->Status(std::string(msg.str()));
        percent[1] = percent[0];
        }

    // Display results
    std::stringstream msg;
    msg << errors << " errors detected during test";
    Log(std::string(msg.str()));
    UI->Status(std::string("Idle"));
}

//---------------------------------------------------------------------------
//  ApplicationIo::WriteRom() --  Write rom using relevant settings
//---------------------------------------------------------------------------

void ApplicationIo::WriteRom()
{
    Module().IdRom().Name(Settings.ModuleName);
    Module().IdRom().Revision(Settings.ModuleRevision);
    Module().Clock().ReferenceCalibrationFactor(Settings.PllCorrection);

    for (unsigned int range = 0; range < Ranges(); ++range)
        {
        for (unsigned int ch = 0; ch < Channels(); ++ch)
            {
            Module().Input().Gain(range, ch, Settings.Gain[range][ch]);
            Module().Input().Offset(range, ch, Settings.Offset[range][ch]);
            }

        Module().Input().Calibrated(range, Settings.Calibrated);
        }

    Module().IdRom().StoreToRom();
}

//---------------------------------------------------------------------------
//  ApplicationIo::ReadRom() --  Read rom and update relevant settings
//---------------------------------------------------------------------------

void ApplicationIo::ReadRom()
{
    Module().IdRom().LoadFromRom();

    Settings.ModuleName = Module().IdRom().Name();
    Settings.ModuleRevision = Module().IdRom().Revision();
    Settings.PllCorrection = static_cast<float>(Module().Clock().ReferenceCalibrationFactor());

    bool calibrated = true;
    for (unsigned int range = 0; range < Ranges(); ++range)
        {
        for (unsigned int ch = 0; ch < Channels(); ++ch)
            {
            Settings.Gain[range][ch] = Module().Input().Gain(range, ch);
            Settings.Offset[range][ch] = Module().Input().Offset(range, ch);
            }

        calibrated &= Module().Input().Calibrated(range);
        }

    Settings.Calibrated = calibrated;
}

//==============================================================================
//  CLASS ApplicationSettings
//==============================================================================

//------------------------------------------------------------------------
//  ApplicationSettings::ApplicationSettings() --  Ctor, load from INI file
//------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings()
    : ActiveChannels(AnalogInChannels()), AlertEnable(NumberOfAlerts()),
      Range(AnalogInChannels())
{
    //  Board Settings
    Install( ToIni("Target",               Target,          0)  );
    Install( ToIni("BusmasterSize",        BusmasterSize,   0)  );
    Install( ToIni("ExoFile",              ExoFile,  std::string("")) );

    //  Config Data
    Install( ToIni("ActiveChannels", "Ch", ActiveChannels,    char(0) ));
    Install( ToIni("PacketSize",           PacketSize,        0x10000u)  );
    Install( ToIni("SampleClockSource",    SampleClockSource, 2)  );
    Install( ToIni("SampleRate",           SampleRate,        static_cast<float>(MaxInRateMHz()*1.0e3))  );
    Install( ToIni("ReferenceSource",      ReferenceSource, 0)  );
    Install( ToIni("ReferenceFrequency",   ReferenceFrequency, 100.0f)  );
    Install( ToIni("ExternalTrigger",      ExternalTrigger,     false)  );
    Install( ToIni("FrontPanelTrigger",    FrontPanelTrigger,   false)  );

    Install( ToIni("SyncMaster",           SyncMaster,          1)  );
    Install( ToIni("Framed",               Framed,              0)  );
    Install( ToIni("FrameSize",            FrameSize,           0x4000)  );
    Install( ToIni("AutoTrigger",          AutoTrigger,         true)  );
    Install( ToIni("AutoStop",             AutoStop,            true)  );
    Install( ToIni("DioConfig",            DioConfig,           0)  );
    Install( ToIni("FrontPanelConfig",     FrontPanelConfig,    0x3)  );
    Install( ToIni("TestCounterEnable",    TestCounterEnable,   false)  );
    Install( ToIni("AlertEnable", "Alert", AlertEnable,         char(0) ));

    Install( ToIni("PretriggerEnable",     PretriggerEnable,    false)  );
    Install( ToIni("PretriggerCount",      PretriggerCount,    0u)  );
    Install( ToIni("TriggerDelay",         TriggerDelay,    0u)  );

    //  Streaming
    Install( ToIni("LoggerEnable",         LoggerEnable,    true)  );
    Install( ToIni("PlotEnable",           PlotEnable,      true)  );
    Install( ToIni("OverwriteBdd",         OverwriteBdd,    true)  );
    Install( ToIni("SamplesToLog",         SamplesToLog,    100000u)  );

    //  Real Time Plotting
    Install( ToIni("RTPlotEnable",         RTPlotEnable,          false)  );
    Install( ToIni("RTPlotFrequency",      RTPlotFrequency,       1.0f)  );
    Install( ToIni("RTPlotSamples",        RTPlotSamples,         16384)  );

    Install( ToIni("Help",                 Help,  true) );

    Install( ToIni("Debug Script",         DebugScript,  std::string("")) );
    Install( ToIni("PhaseDetectorFrequency",PhaseDetectorFrequency, 100000.0) );

    std::string file = Root() + "..\\..\\Startup.fr";
    Install( ToIni("Startup Script",          StartupScript, file) );

    Install( ToIni("Range",                Range, 0) );

    Load();

    //Check Alert size
    if(AlertEnable.size() != (size_t)NumberOfAlerts())
        AlertEnable.resize(NumberOfAlerts());

    //
    //  Sanity Check on Target
    if (Target < 0)
        Target = 0;
    //
    //  Sanity Check on Active Channels
    //   (Make sure at least one active channel)
    // Insure at least one active channel, by default
    unsigned int count = 0;
    for (size_t i = 0; i < ActiveChannels.size(); ++i)
        count += ActiveChannels[i];
    ActiveChannels[0] = !count ? true : ActiveChannels[0];

    for (size_t i = 0; i < AnalogInRanges(); ++i)
        {
        Gain.push_back(IniFile::FloatArray(AnalogInChannels()));
        Offset.push_back(IniFile::FloatArray(AnalogInChannels()));
        }
}

//------------------------------------------------------------------------
//  ApplicationSettings::~ApplicationSettings() --  Dtor, save to INI file
//------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    //
    //  Sanity Check on Target
    if (Target < 0)
        Target = 0;

    Save();
}

//------------------------------------------------------------------------

