#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <Framework/Qt/GuiXfr_Qt.h>
#include <Malibu_Mb.h>
#include <StringSupport_Mb.h>
#include <MalibuVersion_Mb.h>

using namespace Innovative;
using namespace std;


//===========================================================================
//  CLASS MainWindow  -- GUI Form
//===========================================================================

//---------------------------------------------------------------------------
//  MainWindow::MainWindow() --  Ctor
//---------------------------------------------------------------------------

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    ToCaption(this, ModuleNameStr() + std::string(" Snap Example"));

    Glob = new UserInterface(this);
    Io = new ApplicationIo(Glob);

    InstallSettings();

    FiclInit();   // Set up FICL system

    //Information about the imgae
    ToControl(ui->ConfigSwVersionLineEdit, Io->SoftwareVersionString());
    ToControl(ui->ConfigSwBuildDateLineEdit, std::string(__DATE__));
    ToControl(ui->ConfigSwBuildTimeLineEdit, std::string(__TIME__));
    ToControl(ui->ConfigMalibuVersionLineEdit, MalibuVersion().Malibu());

    PerformanceStatusBar = new QStatusBar(ui->PerformancStatusBarFrame);
    PerformanceStatusBar->setSizeGripEnabled(false);
    PeriodicStatus();

    LogicStatusBar = new QStatusBar(ui->LogicStatusBarFrame);
    LogicStatusBar->setSizeGripEnabled(false);
    LogicStatus("Unknown");

    UpdateLogicLoadProgress(0);


    // After open.
    // RangeGrid
    SetDimensions(ui->RangeGrid, 1, Io->Channels());
    ToColumnHeading(ui->RangeGrid, 0, "Mode");
    for (unsigned int i = 0; i < Io->Channels(); ++i)
        ToRowHeading(ui->RangeGrid, i, std::string("Ch")+IntToString(i));

    SetDimensions(ui->CoefficientsGrid, Io->Ranges()*2, Io->Channels());

    for (unsigned int range = 0; range < Io->Ranges(); ++range)
        {
        stringstream gain;
        gain << "Gain[" << range << "]";
        ToColumnHeading(ui->CoefficientsGrid, range*2, gain.str());

        stringstream offset;
        offset << "Offset[" << range << "]";
        ToColumnHeading(ui->CoefficientsGrid, range*2+1, offset.str());
        }

    for (unsigned int i = 0; i < Io->Channels(); ++i)
        {
        std::string tag = std::string("Ch ") + IntToString(i);

        ToRowHeading(ui->CoefficientsGrid, i, tag);
        AddToControl(ui->ActiveChannelsCheckListBox, tag);
        }

    //  Set up Target Listbox
    Io->BBoardCount = Io->BoardCount();
    if (!Io->BBoardCount)
        Log("No devices detected!");

    // Populate target combo box with actual boards detected
    BoardNames = Io->BoardNames();
    FillControl(ui->TargetComboBox, BoardNames);
    ControlEnabled(ui->TargetComboBox, Io->BBoardCount ? true : false);

    SetSettings();

    // Initialize grids, Disable selection region
    ShowTab(ui->ConfigTabSheet, ui->TabSheet, true);

    AdjustEnables();

    ui->TargetComboBox->setCurrentIndex(Io->SetTargetIndex());

    Log(Io->PrintDevices());
}

//---------------------------------------------------------------------------
//  MainWindow::~MainWindow() --  Dtor
//---------------------------------------------------------------------------

MainWindow::~MainWindow()
{
   GetSettings();

    delete Io;
    Io = 0;

    delete Glob;
    Glob = 0;

    delete ui;
}

//---------------------------------------------------------------------------
//  MainWindow::~changeEvent() --
//---------------------------------------------------------------------------

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

//===========================================================================
//  Event Handlers
//===========================================================================

//---------------------------------------------------------------------------
//  MainWindow::on_OpenBtn_clicked() --  Open driver
//---------------------------------------------------------------------------

void MainWindow::on_OpenBtn_clicked()
{
    Io->Settings.Target = ReadControl<int>(ui->TargetComboBox);
//    if (BoardNames[Io->Settings.Target] != "X3 XMCe")
//        {
//        QMessageBox::information(this,tr("Wrong Target"), tr("Target should be X3 XMCe. Please check target."));
//        return;
//        }

    Io->Open();

    AdjustEnables();

    SetSettings();
    on_LoadFromRomButton_clicked();
}

//---------------------------------------------------------------------------
//  MainWindow::on_CloseBtn_clicked() --  Close driver
//---------------------------------------------------------------------------

void MainWindow::on_CloseBtn_clicked()
{
    Io->Close();

    AdjustEnables();
}

//---------------------------------------------------------------------------
//  MainWindow::on_StreamStartBtn_clicked() --  Start stream
//---------------------------------------------------------------------------

void MainWindow::on_StreamStartBtn_clicked()
{
    ClearControl(ui->MessageLog);
    //
    //  Activate stop button, deactivate start btn
    ControlEnabled(ui->StreamStartBtn, false);
    ControlEnabled(ui->StreamStopBtn, true);

    Io->StartStreaming();
}

//---------------------------------------------------------------------------
//  MainWindow::on_StreamStopBtn_clicked() --  Stop stream
//---------------------------------------------------------------------------

void MainWindow::on_StreamStopBtn_clicked()
{
    //
    //  Deectivate stop button, activate start btn
    ControlEnabled(ui->StreamStartBtn, true);
    ControlEnabled(ui->StreamStopBtn, false);

    Io->StopStreaming();
}

//---------------------------------------------------------------------------
//  MainWindow::on_LoadFromRomButton_clicked() --  Load ROM
//---------------------------------------------------------------------------

void MainWindow::on_LoadFromRomButton_clicked()
{
    Io->ReadRom();

    for (unsigned int range = 0; range < Io->Ranges(); ++range)
        {
        for (unsigned int ch = 0; ch < Io->Channels(); ++ch)
            {
            // Gain.
            ToControl(ui->CoefficientsGrid, range*2, ch, Io->Settings.Gain[range][ch], 4);
            // Offset.
            ToControl(ui->CoefficientsGrid, range*2+1, ch, Io->Settings.Offset[range][ch], 4);
            }
        }

    ToControl(ui->CalibratedCheckBox, Io->Settings.Calibrated);

    ToControl(ui->PllCorrectionEdit, Io->Settings.PllCorrection, 4);
    ToControl(ui->PllCalibratedCheckBox, Io->Settings.PllCalibrated);

    ToControl(ui->ModuleNameEdit, Io->Settings.ModuleName);
    ToControl(ui->ModuleRevisionEdit, Io->Settings.ModuleRevision);
 }

//---------------------------------------------------------------------------
//  MainWindow::on_StoreToRomButton_clicked() --  Load ROM
//---------------------------------------------------------------------------

void MainWindow::on_StoreToRomButton_clicked()
{
    for (unsigned int range = 0; range < Io->Ranges(); ++range)
        {
        for (unsigned int ch = 0; ch < Io->Channels(); ++ch)
            {
            FromControl(ui->CoefficientsGrid, range*2, ch, Io->Settings.Gain[range][ch]);
            FromControl(ui->CoefficientsGrid, range*2+1, ch, Io->Settings.Offset[range][ch]);
            }
        }

    FromControl(ui->CalibratedCheckBox, Io->Settings.Calibrated);

    FromControl(ui->PllCorrectionEdit, Io->Settings.PllCorrection);
    FromControl(ui->PllCalibratedCheckBox, Io->Settings.PllCalibrated);
    FromControl(ui->ModuleNameEdit, Io->Settings.ModuleName);
    FromControl(ui->ModuleRevisionEdit, Io->Settings.ModuleRevision);

    Io->WriteRom();
}

//---------------------------------------------------------------------------
//  MainWindow::on_DefaultRomButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_DefaultRomButton_clicked()
{
    for (unsigned int range = 0; range < Io->Ranges(); ++range)
        {
        for (unsigned int ch = 0; ch < Io->Channels(); ++ch)
            {
            // Gain.
            ToControl(ui->CoefficientsGrid, range*2, ch, "1.0000");
            // Offset.
            ToControl(ui->CoefficientsGrid, range*2+1, ch, "0.0000");
            }
        }

    ToControl(ui->CalibratedCheckBox, false);

    ToControl(ui->PllCorrectionEdit, "1.0000");
    ToControl(ui->PllCalibratedCheckBox, false);

    ToControl(ui->ModuleNameEdit, Io->ModuleName());
    ToControl(ui->ModuleRevisionEdit, Io->ModuleRevision());
}

//---------------------------------------------------------------------------
//  MainWindow::on_DebugScriptBrowseButton_clicked() --  Select debug script
//---------------------------------------------------------------------------

void MainWindow::on_DebugScriptBrowseButton_clicked()
{
    SelectFileDialog(ui->DebugScriptEdit, "Script Files (*.fr *.*");
}

//---------------------------------------------------------------------------
//  MainWindow::on_DebugScriptBrowseButton_clicked() --  Execute debug script
//---------------------------------------------------------------------------

void MainWindow::on_DebugScriptExecuteButton_clicked()
{
//    Io->ExecuteDebugScript(ReadControl<std::string>(ui->DebugScriptEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_SoftwareAlertButton_clicked() --  Execute software alert
//---------------------------------------------------------------------------

void MainWindow::on_SoftwareAlertButton_clicked()
{
    Io->SoftwareAlert(ReadControl<int>(ui->SoftwareAlertEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_LogicReadButton_clicked() --  Read logic port
//---------------------------------------------------------------------------

void MainWindow::on_LogicReadButton_clicked()
{
    Io->Logic(ReadControl<std::string>(ui->AddressEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_LogicWriteButton_clicked() --  Write logic port
//---------------------------------------------------------------------------

void MainWindow::on_LogicWriteButton_clicked()
{
    Io->Logic(ReadControl<std::string>(ui->AddressEdit), ReadControl<std::string>(ui->DataEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_PortReadButton_clicked() --  Read PCI port
//---------------------------------------------------------------------------

void MainWindow::on_PortReadButton_clicked()
{
    Io->Port(ReadControl<std::string>(ui->AddressEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_PortWriteButton_clicked() --  Write PCI port
//---------------------------------------------------------------------------

void MainWindow::on_PortWriteButton_clicked()
{
    Io->Port(ReadControl<std::string>(ui->AddressEdit), ReadControl<std::string>(ui->DataEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_DeviceReadButton_clicked() --  Read device port
//---------------------------------------------------------------------------

void MainWindow::on_DeviceReadButton_clicked()
{
    Io->Device(ReadControl<std::string>(ui->AddressEdit),
               ReadControl<std::string>(ui->DeviceEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_DeviceWriteButton_clicked() --  Write device port
//---------------------------------------------------------------------------

void MainWindow::on_DeviceWriteButton_clicked()
{
    Io->Device(ReadControl<std::string>(ui->AddressEdit),
               ReadControl<std::string>(ui->DeviceEdit),
               ReadControl<std::string>(ui->DataEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_ClockSourceComboBox_currentIndexChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_ClockSourceComboBox_currentIndexChanged(int /*index*/)
{
    AdjustEnables();
}


//---------------------------------------------------------------------------
//  MainWindow::on_HelpButton_clicked() --  Display FICL help
//---------------------------------------------------------------------------

void MainWindow::on_HelpButton_clicked()
{
    Io->Help();
}

//---------------------------------------------------------------------------
//  MainWindow::on_ZbtTestButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_ZbtTestButton_clicked()
{
    Io->TestZbt();
}

//---------------------------------------------------------------------------
//  MainWindow::on_LogicLoadBrowseBtn_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_LogicLoadBrowseBtn_clicked()
{
    SelectFileDialog(ui->LogicFilenameEdit, "Image File *.bit *.exo *.*");
}

//---------------------------------------------------------------------------
//  MainWindow::on_LogicLoadConfigBtn_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_LogicLoadConfigBtn_clicked()
{
    Io->LoadLogic();
}

//===========================================================================
//  Support Methods
//===========================================================================

//---------------------------------------------------------------------------
//  MainWindow::AdjustEnables() --
//---------------------------------------------------------------------------

void MainWindow::AdjustEnables()
{
    ControlEnabled(ui->TargetComboBox, !Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->TargetLabel, !Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->BusmasterSizeComboBox, !Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->BusmasterSizeLabel, !Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->OpenBtn, !Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->CloseBtn, Io->IsOpen());
    ControlEnabled(ui->LogicLoadConfigBtn, Io->IsOpen());
    ControlEnabled(ui->ReferenceSourceComboBox, ReadControl<int>(ui->ClockSourceComboBox)==2);
    ControlEnabled(ui->ReferenceSourceLabel, ReadControl<int>(ui->ClockSourceComboBox)==2);

    ControlEnabled(ui->StreamTabSheet, Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->ZbtTabSheet, Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->EepromTabSheet, Io->IsOpen()&&Io->BBoardCount);
    ControlEnabled(ui->DebugTabSheet, Io->IsOpen()&&Io->BBoardCount);

    ControlVisible(ui->PretriggerGroupBox, Io->PretriggerSupported());
    ControlVisible(ui->TriggerDelayGroupBox, Io->TriggerDelaySupported());
}

//---------------------------------------------------------------------------
//  MainWindow::Log() --
//---------------------------------------------------------------------------

void  MainWindow::Log(const std::string & a_string)
{
        AddToControl(ui->MessageLog, a_string);
}

//---------------------------------------------------------------------------
//  MainWindow::LogicStatus() --
//---------------------------------------------------------------------------

void  MainWindow::LogicStatus(const std::string & a_string)
{
    ToControl(LogicStatusBar, 0, a_string);
}


//---------------------------------------------------------------------------
//  MainWindow::Status() --
//---------------------------------------------------------------------------

void  MainWindow::Status(const std::string & a_string)
{
    ToControl(ui->StatusBar, 0, a_string);
}

//---------------------------------------------------------------------------
//  MainWindow::AfterStreamAutoStop() --
//---------------------------------------------------------------------------

void  MainWindow::AfterStreamAutoStop()
{
    //
    //  Deectivate stop button, activate start btn
    ControlEnabled(ui->StreamStartBtn, true);
    ControlEnabled(ui->StreamStopBtn, false);

    // Update the Stream statistics
    PeriodicStatus();
}

//---------------------------------------------------------------------------
//  MainWindow::UpdateLogicLoadProgress() --
//---------------------------------------------------------------------------

void  MainWindow::UpdateLogicLoadProgress(int percent)
{
    ToControl(ui->LogicLoadProgress, percent);
}

//---------------------------------------------------------------------------
//  MainWindow::Warning() --
//---------------------------------------------------------------------------

void MainWindow::Warning(const std::string & msg)
{
    WarningDialog(msg);
}

//---------------------------------------------------------------------------
//  MainWindow::PeriodicStatus() --
//---------------------------------------------------------------------------

void MainWindow::PeriodicStatus()
{
    ToControl(PerformanceStatusBar, 0, Io->BlockCount(), 0, "Block Count");
    ToControl(PerformanceStatusBar, 1, static_cast<float>(Io->BlockRate()), 1, "Rate (MB/s)");
    ToControl(PerformanceStatusBar, 2, Io->Temperature(), 0, "Temp (C)");
    ToControl(PerformanceStatusBar, 3, Io->DioData(), 16, "Dig In");
    if (Io->Settings.SampleClockSource == 2) // If PLL selected
        ToControl(PerformanceStatusBar, 4, (Io->Pll_Locked()) ? "+" : "-", "PLL Locked: ");
    else
        ToControl(PerformanceStatusBar, 4, "N/A", "PLL Locked: ");
}

//---------------------------------------------------------------------------
//  MainWindow::ExternalClockFrequency() --
//---------------------------------------------------------------------------

void  MainWindow::ExternalClockFrequency(float value)
{
    ToControl(ui->InputFrequencyEdit, value, 3);
}

//---------------------------------------------------------------------------
//  MainWindow::FiclInit() --
//---------------------------------------------------------------------------

void  MainWindow::FiclInit()
{
    if (HasFiclSupport()==false)
        return;

    //  Link Module to Ficl System
    Io->FiclInit();
}

//-----------------------------------------------------------------------------
//  MainWindow::FiclReadCommandString()  --
//-----------------------------------------------------------------------------

std::string MainWindow::FiclReadCommandString()
{
    return ReadControl<std::string>(ui->FiclCommandEdit);
}

//-----------------------------------------------------------------------------
//  MainWindow::FiclCommandComplete()  --
//-----------------------------------------------------------------------------

void MainWindow::FiclCommandComplete()
{
    ClearControl(ui->FiclCommandEdit);
}

//---------------------------------------------------------------------------
//  MainWindow::on_FiclCommandEdit_returnPressed() --
//---------------------------------------------------------------------------

void MainWindow::on_FiclCommandEdit_returnPressed()
{
    Io->FiclIo_OnCommandTextComplete();
}

//---------------------------------------------------------------------------
//  MainWindow::InstallSettings() --  Initialize Transfer Object
//---------------------------------------------------------------------------

void MainWindow::InstallSettings()
{
    //  Board Settings
    GL.Install( ToGui(ui->TargetComboBox,        Io->Settings.Target) );
    GL.Install( ToGui(ui->BusmasterSizeComboBox, Io->Settings.BusmasterSize) );
    GL.Install( ToGui(ui->LogicFilenameEdit, Io->Settings.ExoFile) );

    //  Config Data
    GL.Install( ToGui(ui->ActiveChannelsCheckListBox,     Io->Settings.ActiveChannels) );
    GL.Install( ToGui(ui->PacketSizeEdit,                 Io->Settings.PacketSize, 16) );
    GL.Install( ToGui(ui->ClockSourceComboBox,            Io->Settings.SampleClockSource) );
    GL.Install( ToGui(ui->SampleRateEdit,                 Io->Settings.SampleRate, 5) );
    GL.Install( ToGui(ui->ReferenceSourceComboBox,        Io->Settings.ReferenceSource) );
    GL.Install( ToGui(ui->InputFrequencyEdit,             Io->Settings.ReferenceFrequency, 3) );
    GL.Install( ToGui(ui->TriggerSourceRadioGroup,        Io->Settings.ExternalTrigger) );
    GL.Install( ToGui(ui->ExtTriggerSourceRadioGroup,     Io->Settings.FrontPanelTrigger) );
    GL.Install( ToGui(ui->TriggerModeRadioGroup,          Io->Settings.Framed) );
    GL.Install( ToGui(ui->SyncMasterRadioGroup,           Io->Settings.SyncMaster) );
    GL.Install( ToGui(ui->FrameSizeEdit,                  Io->Settings.FrameSize, 16) );
    GL.Install( ToGui(ui->AutoTriggerCheckBox,            Io->Settings.AutoTrigger) );
    GL.Install( ToGui(ui->AutoStopCheckBox,               Io->Settings.AutoStop) );
    GL.Install( ToGui(ui->DioConfigEdit,                  Io->Settings.DioConfig, 16) );
    GL.Install( ToGui(ui->FrontPanelConfigEdit,           Io->Settings.FrontPanelConfig, 16) );
    GL.Install( ToGui(ui->TestCounterEnableCheckBox,      Io->Settings.TestCounterEnable) );
    GL.Install( ToGui(ui->TestCounterEnableCheckBox,      Io->Settings.TestCounterEnable) );
    GL.Install( ToGui(ui->AlertEnableCheckListBox,        Io->Settings.AlertEnable) );

    GL.Install( ToGui(ui->PretriggerEnableChk,            Io->Settings.PretriggerEnable) );
    GL.Install( ToGui(ui->PretriggerCountEdit,            Io->Settings.PretriggerCount) );
    GL.Install( ToGui(ui->TriggerDelayEdit,               Io->Settings.TriggerDelay) );

    //  Streaming
    GL.Install( ToGui(ui->LoggerEnableCheckBox, Io->Settings.LoggerEnable) );
    GL.Install( ToGui(ui->PlotEnableCheckBox, Io->Settings.PlotEnable) );
    GL.Install( ToGui(ui->OverwriteBddCheckBox, Io->Settings.OverwriteBdd) );
    GL.Install( ToGui(ui->SamplesEdit, Io->Settings.SamplesToLog) );

    //  Real Time Plotting
    GL.Install( ToGui(ui->RTPlotEnableCheckBox, Io->Settings.RTPlotEnable) );
    GL.Install( ToGui(ui->RTPlotFrequencyEdit, Io->Settings.RTPlotFrequency) );
    GL.Install( ToGui(ui->RTPlotSamplesEdit, Io->Settings.RTPlotSamples, 16) );

    // Debug
    GL.Install( ToGui(ui->DebugScriptEdit, Io->Settings.DebugScript) );
}

//---------------------------------------------------------------------------
//  MainWindow::GetSettings() --  Form controls to options object
//---------------------------------------------------------------------------

void MainWindow::GetSettings()
{
    for (unsigned int i = 0; i < Io->Channels(); ++i)
        FromControl(ui->RangeGrid, 0, i, Io->Settings.Range[i]);

    ControlVisible(ui->groupBox_7, HasRangeGrid());

    GL.GetSettings();    // alternate name
}

//---------------------------------------------------------------------------
//  MainWindow::SetSettings() --  Settings object to form controls
//---------------------------------------------------------------------------

void MainWindow::SetSettings()
{
    GL.SetSettings();

    for (unsigned int i = 0; i < Io->Channels(); ++i)
        ToControl(ui->RangeGrid, 0, i, Io->Settings.Range[i]);
}

//---------------------------------------------------------------------------
//  MainWindow::CalculateExternalFrequency() --
//---------------------------------------------------------------------------

void MainWindow::on_CalculatExtClockRateBtn_clicked()
{
    Io->CalculateExternalFrequency();
}
