// ApplicationIo.cpp
//
// Board-specific data flow and hardware I/O
// Copyright 2006 Innovative Integration
//---------------------------------------------------------------------------

#define NOMINMAX

#include <Malibu_Mb.h>
#include <Analysis_Mb.h>
#include <IppMemoryUtils_Mb.h>

#include "ApplicationIo.h"
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <FileSupport_Mb.h>
#include <BinView_Mb.h>
#include <sstream>
#include <fstream>
#include <cmath>
#include <Buffer_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <BufferHeader_Mb.h>
#include <Magic_Mb.h>
#include <algorithm>
#include <Application/StaticWaveform_App.h>
#include <Exception_Mb.h>

using namespace Innovative;
using namespace InnovativeKernel;
using namespace std;

const int HeaderTagValuePostPacketizer = 0x00000000;
const int HeaderTagValueOriginal = HeaderTagValuePostPacketizer;

//===========================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    : Settings(NumChannels), UI(ui), Time(4),
      FOpened(false), FStreaming(false), FStreamConnected(false),
      FBlockCount(0)
{
    TraceVerbosity(Trace::vNormal);

    OnLog.SetEvent(this, &ApplicationIo::HandleOnLog);

    //
    // Use IPP performance memory functions.
    Init::UsePerformanceMemoryFunctions();

    // Hook script event handlers.
    Script.OnCommand.SetEvent(this, &ApplicationIo::HandleScriptCommand);
    Script.OnMessage.SetEvent(this, &ApplicationIo::HandleScriptMessage);
    //
    //  Configure Module Event Handlers
    Module.OnBeforeStreamStart.SetEvent(this, &ApplicationIo::HandleBeforeStreamStart);
    Module.OnBeforeStreamStart.Synchronize();
    Module.OnAfterStreamStart.SetEvent(this, &ApplicationIo::HandleAfterStreamStart);
    Module.OnAfterStreamStart.Synchronize();
    Module.OnAfterStreamStop.SetEvent(this, &ApplicationIo::HandleAfterStreamStop);
    Module.OnAfterStreamStop.Synchronize();

    //
    //  Alerts
    Module.Alerts().OnTimeStampRolloverAlert.SetEvent(this, &ApplicationIo::HandleTimestampRolloverAlert);
    Module.Alerts().OnSoftwareAlert.SetEvent(this, &ApplicationIo::HandleSoftwareAlert);
    Module.Alerts().OnWarningTemperature.SetEvent(this, &ApplicationIo::HandleWarningTempAlert);
    Module.Alerts().OnOutputFifoUnderrun.SetEvent(this, &ApplicationIo::HandleOutputFifoUnderrunAlert);
    Module.Alerts().OnOutputTrigger.SetEvent(this, &ApplicationIo::HandleOutputTriggerAlert);
    Module.Alerts().OnPllLost.SetEvent(this, &ApplicationIo::HandlePllLostAlert);

    //
    //  Configure Module Event Handlers
    Module.Logic().OnFpgaFileReadProgress.SetEvent(this, &ApplicationIo::HandleProgress);
    Module.Logic().OnFpgaFileReadComplete.SetEvent(this, &ApplicationIo::HandleParseComplete);
    Module.Logic().OnFpgaParseProgress.SetEvent(this, &ApplicationIo::HandleProgress);
    Module.Logic().OnFpgaParseComplete.SetEvent(this, &ApplicationIo::HandleParseComplete);
    Module.Logic().OnFpgaParseMessage.SetEvent(this, &ApplicationIo::HandleLoadError);
    Module.Logic().OnFpgaLoadProgress.SetEvent(this, &ApplicationIo::HandleProgress);
    Module.Logic().OnFpgaLoadComplete.SetEvent(this, &ApplicationIo::HandleLoadComplete);
    Module.Logic().OnFpgaLoadMessage.SetEvent(this, &ApplicationIo::HandleLoadError);
    //
    //  Configure Stream Event Handlers
    Stream.OnDataRequired.SetEvent(this, &ApplicationIo::HandleDataRequired);

    Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);
    Timer.OnElapsed.Thunk();

    //
    //  Set up Loggers and graphs
    std::stringstream ss;
    ss << Settings.Root() << "Data.bin";

    Timer.Interval(1000);

    std::stringstream msg;
    msg << "Be sure to read the help file for info on this program";
    msg << " located in the root of the this example folder.  ";
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

int ApplicationIo::BoardCount()
{
    return static_cast<int>(Module.BoardCount());
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardNames() -- Get Device Names
//---------------------------------------------------------------------------

vector<string> ApplicationIo::BoardNames()
{
    return Module.BoardNames();
}

//---------------------------------------------------------------------------
//  ApplicationIo::SetTargetIndex() -- Set Target Index
//---------------------------------------------------------------------------

unsigned int ApplicationIo::SetTargetIndex()
{
    vector<string> BNames = BoardNames();
    for (size_t i = 0; i < BNames.size(); i++)
        {
        string Name = BNames[i];
        size_t found = Name.find("X3 XMCe");
        if (found != string::npos)
            {
            return i;
            }
        }

    return 0;
}

//---------------------------------------------------------------------------
//  ApplicationIo::PrintDevices() -- Print Device Names and Target #
//---------------------------------------------------------------------------

string ApplicationIo::PrintDevices()
{
    return Module.PrintAllDevices();
}

//---------------------------------------------------------------------------
// ApplicationIo::Open()
//---------------------------------------------------------------------------

void ApplicationIo::Open()
{
    UI->GetSettings();
    
    Module.Clock().SetPhaseDetectorFrequency(Settings.PhaseDetectorFrequency);

    //
    //  Open Devices
    FBusmasterSize = 1 << (Settings.BusmasterSize + 22);
    const int Meg = 1024 * 1024;
	Module.IncomingBusMasterSize(1 * Meg);
    Module.OutgoingBusMasterSize(FBusmasterSize);
    Module.Target(Settings.Target);
    
    try
        {
        Module.Open();
        }
     catch (Innovative::MalibuException & exception)
        {
        UI->Status("Module Device Open Failure!");
        UI->Log("Open Failure:");
        UI->Log(exception.what());
        return;
        }
     catch(...)
        {
        UI->Status("Module Device Open Failure!");
        return;
        }

    Module.Reset();
    UI->Status("Module Device Opened...");
	FOpened = true;

	//
	//  Connect Stream
	Stream.ConnectTo(&Module);
	FStreamConnected = true;
	UI->Status("Stream Connected...");

	PrefillPacketCount = Stream.PrefillPacketCount();

	DisplayLogicVersion();

	FChannels = Module.Output().Info().Channels().Channels();

	// Support for fast clock patch
	Module.FastClock(Settings.FastClock);
}

//---------------------------------------------------------------------------
// ApplicationIo::Close()
//---------------------------------------------------------------------------

void ApplicationIo::Close()
{
	Stream.Disconnect();
	Module.Close();
	FStreamConnected = false;
	FOpened = false;
	UI->Status("Stream Disconnected...");
}

//---------------------------------------------------------------------------
// ApplicationIo::StartStreaming()
//---------------------------------------------------------------------------

bool ApplicationIo::StartStreaming()
{
    //
    //  Set up Parameters for Data Streaming
    //  ...First have UI get settings into our settings store
    UI->GetSettings();

    if (!FStreamConnected)
        {
        Log("Stream not connected! -- Open the boards");
        return false;
        }

	//
    // Make sure packets fit nicely in BM region.
	if (FBusmasterSize/4 < (unsigned int)Settings.StreamPacketSize)
        {
        Log("Error: Packet size is larger than recommended size");
        return false;
        }


    if (SampleRate() > Module.Output().Info().MaxRate())
        {
        Log("Sample rate too high.");
        StopStreaming();
        UI->AfterStreamStop();
        return false;
        }

	// Default reference correction factor and rate
	// Set reference based on clock source to obtain correct FrequencyActual
    Module.Clock().ReferenceCalibrationFactor(1.0f);
    double reference = ReferenceRate();

	// Route clock to active analog devices
	switch (Settings.SampleClockSource)
		{
		case 0:	// J1
			Module.Clock().OutputClock(PmcModule::Timebase::oExternal);
			Module.ExternalClockSource(PmcModule::esPanel);
			break;
		case 1: // P16
			Module.Clock().OutputClock(PmcModule::Timebase::oExternal);
			Module.ExternalClockSource(PmcModule::esPxi);
			break;
		case 2: // PLL
            {
			Module.Clock().OutputClock(PmcModule::Timebase::oVco);

            if (Settings.ReferenceSource == 0)
                {
    			reference = Module.Input().Info().ReferenceClock();
                // Apply timebase correction factor, if available
                double correction = Settings.PllCorrection;
                if (correction != correction)
                    correction = 1.0;	// NaN, so fix it
            	Module.Clock().ReferenceCalibrationFactor(correction);
				}
            }
			break;
		}

	Module.Clock().Reference(reference);
	// Route reference.
	PmcModule::IIReferenceSource ref[] = { PmcModule::rsCrystal, PmcModule::rsPxi };
	Module.ReferenceSource(ref[Settings.ReferenceSource]);
	Module.Clock().Frequency(SampleRate());

	// Disable triggering initially
	Module.Output().ExternalTrigger(false);
	Module.Trigger(PmcModule::tOutput, false);

	// Set sync source
	Module.Master(true);

	FBlockCount = 0;
	FBlockRate = 0;
	FTriggered = -1;
	TestCounter = 0;
	Time.Reset();

	PrefillCount = std::max<int>(Settings.PrefillPeriod, 1);

	Module.Dio().DioPortConfig(Settings.DioConfig);
	Module.FrontPanel().FrontPanelPortConfig(Settings.FrontPanelConfig);

	//
	//  Channel Enables
	Module.Input().Info().Channels().DisableAll();
	Module.Output().Info().Channels().DisableAll();
	for (int i = 0; i < Channels(); ++i)
		if ((Settings.ActiveChannels[i] ? true : false))
			Module.Output().Info().Channels().Enabled(i, true);

	int ActiveChannels = Module.Output().Info().Channels().ActiveChannels();
	if (!ActiveChannels)
		{
		Log("Error: Must enable at least one channel");
		UI->AfterStreamStop();
		return false;
		}

	// Interpolation not supported, emit warning
	if (Module.Clock().Decimation() > 1)
		{
		stringstream msg;
		msg << "Interpolation factor " << Module.Clock().Decimation()
			<< " required applied to reach the requested sampling frequency."
			<< " Not currently supported.";
		Log(msg.str());
		UI->AfterStreamStop();
		return false;
		}

	FStreaming = true;

	// Frame count in units of packet elements
	if (Settings.Framed)
		Module.Output().Framed(Settings.FrameCount);
	else
		Module.Output().Unframed();

	enum IUsesX3Alerts::AlertType Alert[] =
		 {
		 IUsesX3Alerts::alertTimeStampRollover,
         IUsesX3Alerts::alertSoftware,
         IUsesX3Alerts::alertWarningTemperature,
         IUsesX3Alerts::alertPllLost,
         IUsesX3Alerts::alertOutputFifoUnderrun,
         IUsesX3Alerts::alertOutputTrigger
         };

    for (unsigned int i = 0; i < Settings.AlertEnable.size(); ++i)
        Module.Alerts().AlertEnable(Alert[i], Settings.AlertEnable[i] ? true : false);

    // Disable prefill if in test mode
    Stream.PrefillPacketCount(PrefillPacketCount);

	// Calculate waveform buffer
    ShortDG Packet_DG(WaveformPacket);
    //
    // Calculate Packet Size in shorts
    int packet_size_shorts = Settings.StreamPacketSize*ActiveChannels;
    while ((packet_size_shorts%4)!=0)
        packet_size_shorts += ActiveChannels;

    Packet_DG.Resize(packet_size_shorts);

    PacketBufferHeader PktBufferHdr(WaveformPacket);
    PktBufferHdr.PacketSize(Settings.StreamPacketSize);
	PktBufferHdr.PeripheralId(Module.Output().PacketId());
	PktBufferHdr[1] = HeaderTagValueOriginal;

	Module.TestCounterEnable(Settings.TestModeEnable);

	//
	//  Builds a one or 2 channel buffer
	BuildWave(WaveformPacket, Settings.WaveType);

	stringstream msg;
	msg.precision(6);
	double actual = Module.Clock().FrequencyActual();
    msg << "Actual sampling rate: " << actual/(1.e3 * Module.Clock().Decimation()) << " KHz";
    Log(msg.str());

    //  Start Streaming
    Stream.Start();
    Log("Stream Mode started");

    FTicks = 0;

    return true;
}

//---------------------------------------------------------------------------
// ApplicationIo::StopStreaming()
//---------------------------------------------------------------------------

void ApplicationIo::StopStreaming()
{
    if (!FStreaming)
        return;

    if (!FStreamConnected)
        {
        Log("Stream not connected! -- Open the boards");
        return;
        }

    //
    //  Stop Streaming
    Stream.Stop();
    FStreaming = false;
    Timer.Enabled(false);

}

//---------------------------------------------------------------------------
// ApplicationIo::SampleRate()
//---------------------------------------------------------------------------

float ApplicationIo::SampleRate() const
{
    return static_cast<float>(Settings.SampleRate*1.e3);
}

//---------------------------------------------------------------------------
// ApplicationIo::ReferenceRate()
//---------------------------------------------------------------------------

float ApplicationIo::ReferenceRate() const
{
    return static_cast<float>(Settings.ReferenceFrequency*1.e6);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Support Functions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//---------------------------------------------------------------------------
//  ApplicationIo::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

float ApplicationIo::Temperature()
{
    return Module.Thermal().LogicTemperature();
}

//---------------------------------------------------------------------------
//  ApplicationIo::DioData() --  Set DIO bits state
//---------------------------------------------------------------------------

void ApplicationIo::DioData(unsigned int value)
{
    return Module.Dio().DioPortData().Value(value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DioData() --  Query DIO bits state
//---------------------------------------------------------------------------

unsigned int ApplicationIo::DioData()
{
    return Module.Dio().DioPortData().Value();
}

//---------------------------------------------------------------------------
//  ApplicationIo::FrontPanelData() --  Set DIO bits state
//---------------------------------------------------------------------------

void ApplicationIo::FrontPanelData(unsigned int value)
{
    return Module.FrontPanel().FrontPanelPortData().Value(value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::FrontPanelData() --  Query DIO bits state
//---------------------------------------------------------------------------

unsigned int ApplicationIo::FrontPanelData()
{
    return Module.FrontPanel().FrontPanelPortData().Value() & 0xfff;
}

//---------------------------------------------------------------------------
//  ApplicationIo::SoftwareTrigger() --  Issue single software trigger pulse
//------------------------------------------------------------------------------

void  ApplicationIo::SoftwareTrigger()
{
    Module.Trigger(PmcModule::tOutput, false);
    Module.Trigger(PmcModule::tOutput, true);
    ++FTriggered;
}

//------------------------------------------------------------------------------
//  ApplicationIo::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  ApplicationIo::DisplayLogicVersion()
{
    std::stringstream msg;
    msg << std::hex << "Version: " << Module.UserLogicVersion().FpgaLogicVersion()
        << ", Variant: " << Module.UserLogicVersion().FpgaLogicVariant()
        << ", Revision: " << Module.UserLogicVersion().FpgaLogicRevision()
        << ", Type: " << Module.UserLogicVersion().FpgaLogicType();
    UI->LogicVersionStatus(msg.str());
}

//------------------------------------------------------------------------------
//  ApplicationIo::Elapsed() --  Display timestamp as elapsed MCLKs
//------------------------------------------------------------------------------

std::string ApplicationIo::Elapsed(size_t timestamp)
{
    stringstream msg;
    msg << timestamp << " master clocks";
    return msg.str();
}

//---------------------------------------------------------------------------
// ApplicationIo::WriteRom()
//---------------------------------------------------------------------------

void ApplicationIo::WriteRom()
{
    Module.IdRom().Name(Settings.ModuleName);
    Module.IdRom().Revision(Settings.ModuleRevision);
    for (int ch = 0; ch < Channels(); ++ch)
        {
        Module.Output().Gain(ch, Settings.DacGain[ch]);
        Module.Output().Offset(ch, Settings.DacOffset[ch]);
        }

    Module.Output().Calibrated(Settings.Calibrated);
    Module.IdRom().StoreToRom();
}

//---------------------------------------------------------------------------
// ApplicationIo::ReadRom()
//---------------------------------------------------------------------------

void ApplicationIo::ReadRom()
{
    Module.IdRom().LoadFromRom();

    Settings.ModuleName = Module.IdRom().Name();
    Settings.ModuleRevision = Module.IdRom().Revision();

    for (int ch = 0; ch < Channels(); ++ch)
        {
        Settings.DacGain[ch] = Module.Output().Gain(ch);
        Settings.DacOffset[ch] = Module.Output().Offset(ch);
        }
    Settings.Calibrated = Module.Output().Calibrated();
}

//---------------------------------------------------------------------------
// ApplicationIo::IsOpen()
//---------------------------------------------------------------------------

bool ApplicationIo::IsOpen()
{
    return FOpened;
}

//---------------------------------------------------------------------------
// ApplicationIo::IsTriggered()
//---------------------------------------------------------------------------

bool ApplicationIo::IsTriggered()
{
    return FTriggered >= 0;
}

//---------------------------------------------------------------------------
// ApplicationIo::IsStreaming()
//---------------------------------------------------------------------------

bool ApplicationIo::IsStreaming()
{
    return FStreaming;
}

//---------------------------------------------------------------------------
// ApplicationIo::IsDataCheckError()
//---------------------------------------------------------------------------

bool ApplicationIo::IsDataCheckError()
{
    bool status = false;

    return status;
}

//---------------------------------------------------------------------------
// ApplicationIo::IsPllLocked()
//---------------------------------------------------------------------------

bool ApplicationIo::IsPllLocked()
{
	return Module.Clock().Locked();

}

//---------------------------------------------------------------------------
// ApplicationIo::Channels()
//---------------------------------------------------------------------------

int ApplicationIo::Channels() const
{
    return NumChannels;
}

//---------------------------------------------------------------------------
// ApplicationIo::BlockRate() -- Return Tx block rate.
//---------------------------------------------------------------------------

double ApplicationIo::BlockRate() const
{
    return FBlockRate;
}

//------------------------------------------------------------------------
// ApplicationIo::BlockCount() -- Returns Tx block count.
//------------------------------------------------------------------------

int	ApplicationIo::BlockCount() const
{
    return FBlockCount;
}

//---------------------------------------------------------------------------
//  ApplicationIo::DoLogicStore() --  Store to logic register
//---------------------------------------------------------------------------

void ApplicationIo::DoLogicStore()
{
    int addr = Script.Pop();
    int value = Script.Pop();

    PokeLogic(addr, value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DoLogicFetch() --  Fetch from logic register
//---------------------------------------------------------------------------

void ApplicationIo::DoLogicFetch()
{
    int addr = Script.Pop();
    PeekLogic(addr);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DoPortStore() --  Store to PCI register
//---------------------------------------------------------------------------

void ApplicationIo::DoPortStore()
{
    int addr = Script.Pop();
    int value = Script.Pop();
    PokePort(addr, value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DoPortFetch() --  Fetch from PCI register
//---------------------------------------------------------------------------

void ApplicationIo::DoPortFetch()
{
    int addr = Script.Pop();
    PeekPort(addr);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DoDelay() --  Spin for a while
//---------------------------------------------------------------------------

void ApplicationIo::DoDelay()
{
    int value = Script.Pop();
    Innovative::Sleep(value);

    std::stringstream msg;
    msg << " Sleep " << value << " mS";
    Log(std::string(msg.str()));
}

//------------------------------------------------------------------------------
// ApplicationIo::SendOneBlock()
//------------------------------------------------------------------------------

void ApplicationIo::SendOneBlock(PacketStream * PS)
{
    ShortDG Packet_DG(WaveformPacket);

    // Calculate transfer rate in kB/s
    double Period = Time.Differential();
    if (Period)
        FBlockRate = Packet_DG.SizeInBytes() / (Period*1.0e3);

    //
    //  No matter what channels are enabled, we have one packet type
    //    to send here
    PS->Send(WaveformPacket);

    ++FBlockCount;
}

//------------------------------------------------------------------------
//  ApplicationIo::BuildWave() -- File buffer with specified waveform
//------------------------------------------------------------------------

void ApplicationIo::BuildWave(Buffer & Buffer, int WaveType)
{
    StaticWaveform  Wave;
    WaveformFile    WaveFile;

    Wave.Frequency(Settings.WaveformFrequency * 1.e3);
    Wave.SampleRate(Module.Clock().FrequencyActual());
    Wave.Amplitude(Settings.WaveformAmplitude);
    unsigned int channels = Module.Output().Info().Channels().ActiveChannels();
    unsigned int bits = Module.Output().Info().Channels().Bits();
    unsigned int samples = static_cast<unsigned int>(Buffer.SizeInBytes())/Module.Output().Info().Channels().BytesPerEvent();
    Wave.Type(static_cast<StaticWaveform::IIType>(WaveType));
    Wave.Format(channels, bits, samples);
    Wave.Accumulate();

    WaveFile.SampleRate(SampleRate());
    WaveFile.Format(channels, bits, samples);

    if (WaveType != 4)  //wtFile
        Wave.Apply(Buffer);
    else
        WaveFile.Load(Settings.WaveFile, Buffer);

    WaveFile.Save(Settings.Root() + string("BuildWave.bin"), Buffer);
}

//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Configuration Event Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  ApplicationIo::HandleScriptMessage() --  Log script message
//---------------------------------------------------------------------------

void ApplicationIo::HandleScriptMessage(ProcessStatusEvent & Event)
{
    Log(Event.Message);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleScriptCommand() --  Dispatch script command
//---------------------------------------------------------------------------

void ApplicationIo::HandleScriptCommand(Innovative::ProcessStatusEvent & Event)
{
    const std::string & cmd = Event.Message;

    // Logic store
    if (cmd == "l!")
        DoLogicStore();
    // Logic fetch
    if (cmd == "l@")
        DoLogicFetch();
    // Port store
    if (cmd == "p!")
        DoPortStore();
    // Port fetch
    if (cmd == "p@")
        DoPortFetch();
    // Dwell
    if (cmd == "ms")
        DoDelay();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Data Flow Event Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  ApplicationIo::HandleBeforeStreamStart() --  Pre-streaming init event
//---------------------------------------------------------------------------

void ApplicationIo::HandleBeforeStreamStart(OpenWire::NotifyEvent & /*Event*/)
{
    if (Settings.ScriptEnable[0])
        {
        Log(std::string("Running BSS script..."));
        Script.Execute(Settings.BssScript);
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleAfterStreamStart() --  Post streaming init event
//---------------------------------------------------------------------------

void ApplicationIo::HandleAfterStreamStart(OpenWire::NotifyEvent & /*Event*/)
{
    Log(std::string("Analog I/O started"));

    if (Settings.ScriptEnable[1])
        {
        Log(std::string("Running ASS script..."));
        Script.Execute(Settings.AssScript);
        }

    Timer.Enabled(true);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleAfterStreamStop() --  Post stream termination event
//---------------------------------------------------------------------------

void ApplicationIo::HandleAfterStreamStop(OpenWire::NotifyEvent & /*Event*/)
{
    Module.Trigger(PmcModule::tOutput, false);
    Module.Output().ExternalTrigger(false);

    Log(std::string("Analog I/O Stopped"));
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOnLog() --
//---------------------------------------------------------------------------

void ApplicationIo::HandleOnLog(Innovative::ProcessStatusEvent & Event)
{
    UI->Log(Event.Message);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*Event*/)
{
    int DigIn = DioData();
    int FrontIn = FrontPanelData();

    // Display status
    UI->PeriodicStatus();

    FrontPanelData(~FrontIn);
    DioData(~DigIn);

    // Initial trigger state machine below
    if (IsTriggered() || !Settings.AutoTrigger)
        return;

    if (PrefillCount)
        --PrefillCount;

    if ((Settings.ExternalTrigger == 0) && (PrefillCount == 0))
        SoftwareTrigger();
    else
        Module.Output().ExternalTrigger(true);
}

//---------------------------------------------------------------------------
// ApplicationIo::HandleDataRequired()
//---------------------------------------------------------------------------

void ApplicationIo::HandleDataRequired(PacketStreamDataEvent & Event)
{
    SendOneBlock(Event.Sender);
}

//---------------------------------------------------------------------------
//  ApplicationIo::SoftwareAlert() --  Issue specified software alert
//---------------------------------------------------------------------------

void ApplicationIo::SoftwareAlert(unsigned int value)
{
    Log(std::string("Posting SW alert..."));
    Module.Alerts().SoftwareAlert(value);
}

//---------------------------------------------------------------------------
//  ApplicationIo::PeekPort() --  Read a register in Port Momory, Log results
//---------------------------------------------------------------------------

void ApplicationIo::PeekPort(int address)
{
    int value = Module.PeekPort(address);

    std::stringstream msg;
    msg << std::hex << "0x" << address << " p@=" << "0x" << value;
    Log(std::string(msg.str()));
}

//---------------------------------------------------------------------------
//  ApplicationIo::PokePort() --  Write a register in Port Momory, Log results
//---------------------------------------------------------------------------

void ApplicationIo::PokePort(int address, int data)
{
    std::stringstream msg;
    msg << std::hex << "0x" << data << " 0x" << address << " p!";
    Log(std::string(msg.str()));

    Module.PokePort(address, data);
}

//---------------------------------------------------------------------------
//  ApplicationIo::PeekLogic() --  Read a register in Logic memory
//---------------------------------------------------------------------------

void ApplicationIo::PeekLogic(int address)
{
    int value = Module.PeekLogic(address);

    std::stringstream msg;
    msg << std::hex << "0x" << address << " l@=" << "0x" << value;
    Log(std::string(msg.str()));
}

//---------------------------------------------------------------------------
//  ApplicationIo::PokeLogic() --  Write a register in Logic Memory
//---------------------------------------------------------------------------

void ApplicationIo::PokeLogic(int address, int data)
{
    std::stringstream msg;
    msg << std::hex << "0x" << data << " 0x" << address << " l!";
    Log(std::string(msg.str()));

    Module.PokeLogic(address, data);
}

//---------------------------------------------------------------------------
//  ApplicationIo::SpiDump() --  Dump SPI registers
//---------------------------------------------------------------------------

void ApplicationIo::SpiDump()
{

}

//---------------------------------------------------------------------------
//  ApplicationIo::ExecuteDebugScript() --  Run debug script
//---------------------------------------------------------------------------

void ApplicationIo::ExecuteDebugScript()
{
    UI->GetSettings();

    Log(std::string("Running debug script..."));
    Script.Execute(Settings.DebugScript);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Alert Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimestampRolloverAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleTimestampRolloverAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Time stamp rollover " << event.Argument
        << " after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleSoftwareAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleSoftwareAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Software alert 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleWarningTempAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleWarningTempAlert(Innovative::AlertSignalEvent & event)
{
    // Clear warning condition (reading clears)
//    Module.Thermal().LogicWarningTemperature();

    std::stringstream msg;
    msg << "Temp warning alert 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);

    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOutputFifoUnderrunAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleOutputFifoUnderrunAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Output FIFO underrun after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOutputTriggerAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandleOutputTriggerAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Analog triggered after " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandlePllLostAlert() --
//------------------------------------------------------------------------------

void  ApplicationIo::HandlePllLostAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "PLL lock lost at " << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::LoadLogic() -- Initiate Logic Load Process
//---------------------------------------------------------------------------

void  ApplicationIo::LoadLogic()
{
    if (!FOpened)
    {
        Log("No module on specified target");
        return;
    }

    Log("-----------------------------------------------");
    Log("Downloading firmware logic file to card.");
    Log("Please wait ...");

    UI->GetSettings();
    Module.Logic().ConfigureFpga(Settings.ExoFile);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleParseComplete() --  Incremental logic parse event
//------------------------------------------------------------------------------

void  ApplicationIo::HandleParseComplete(ProcessCompletionEvent & /*Event*/)
{
    Log("Parse completed ok");
}

//--------------------------------------------------------------------------
//  ApplicationIo::HandleProgress() --  Incremental logic load event
//--------------------------------------------------------------------------

void  ApplicationIo::HandleProgress(ProcessProgressEvent & event)
{
    UI->UpdateLogicLoadProgress(event.Percent);
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadComplete() --  Logic load completion event
//------------------------------------------------------------------------------

void  ApplicationIo::HandleLoadComplete(ProcessCompletionEvent & /*Event*/)
{
    Log("Load completed ok");

    DisplayLogicVersion();
}

//------------------------------------------------------------------------------
//  ApplicationIo::HandleLoadError() --  Logic load failure event
//------------------------------------------------------------------------------

void ApplicationIo::HandleLoadError(ProcessStatusEvent & event)
{
    Log(event.Message);
}

//===========================================================================
//  CLASS ApplicationSettings  --
//===========================================================================
//------------------------------------------------------------------------
//  ApplicationSettings::ApplicationSettings() --  Ctor, load from INI file
//------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings(int channels)
    : ActiveChannels(channels), ScriptEnable(2), DacGain(channels),
      DacOffset(channels), AlertEnable(6)//, TestFrequency(2)
{
    Application App;
    // Locate files on root of C if remote debugging
    root = (App.Path()[0] == '\\') ? std::string("\\") : App.Path();

    // Locate INI file on root of C if remote debugging
    Ini = new IniFile(root + std::string("Settings.ini"));

    Target  = Ini->Read("Target", 0);
    if (Target < 0)
        Target = 0;

    Ini->Read("ActiveChannels", ActiveChannels);
    // Insure at least one active channel, by default
    int count = 0;
    for (unsigned int i = 0; i < ActiveChannels.size(); ++i)
        count += ActiveChannels[i];
    ActiveChannels[0] = !count ? true : ActiveChannels[0];

    SampleRate        = Ini->Read("SampleRate", Pmc_OutputMaxRatekHz);
    SampleClockSource = Ini->Read("SampleClockSource", 2);
 	ReferenceSource   = Ini->Read("ReferenceSource", 0);
	ReferenceFrequency = Ini->Read("ReferenceFrequency", 100.0f);
    BusmasterSize     = Ini->Read("BusmasterSize", 0);
    ExternalTrigger   = Ini->Read("ExternalTrigger", 0);
    Framed            = Ini->Read("Framed", 0);
    FrameCount        = Ini->Read("FrameCount", 0x4000);
    AutoTrigger       = Ini->Read("AutoTrigger", true);
    BssScript         = Ini->Read("BssScript", std::string(""));
    AssScript         = Ini->Read("AssScript", std::string(""));
    DebugScript       = Ini->Read("DebugScript", std::string(""));
    PhaseDetectorFrequency = Ini->Read("PhaseDetectorFrequency", 100000.0);
    WaveFile          = Ini->Read("WaveFile", std::string(""));
    WaveType          = Ini->Read("WaveType", 1);
    WaveformFrequency = Ini->Read("WaveformFrequency", 1.0f);
	WaveformAmplitude = Ini->Read("WaveformAmplitude", 95.0f);
	DioConfig         = Ini->Read("DioConfig", 0x3f);
    FrontPanelConfig  = Ini->Read("FrontPanelConfig", 0x3);
	PllCorrection     = Ini->Read("PllCorrection", 1.0f);
	PllCalibrated     = Ini->Read("PllCalibrated", false);

	StreamPacketSize  = Ini->Read("StreamPacketSize", 0x10000);

	PrefillPeriod     = Ini->Read("PrefillPeriod", 1);

	Ini->Read("ScriptEnable", ScriptEnable);
	Ini->Read("AlertEnable", AlertEnable);
	ExoFile = Ini->Read("ExoFile", std::string(""));

	PllCorrection = Ini->Read("PllCorrection", 1.0f);
	PllCalibrated = Ini->Read("PllCalibrated", false);

	TestModeEnable  = Ini->Read("TestModeEnable", false);
	FastClock  = Ini->Read("FastClock", true);
}

//------------------------------------------------------------------------
//  ApplicationSettings::~ApplicationSettings() --  Dtor, save to INI file
//------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    if (Target < 0)
        Target = 0;
    Ini->Write("Target", Target);
    Ini->Write("ActiveChannels", ActiveChannels);
    Ini->Write("AlertEnable", AlertEnable);
    Ini->Write("SampleRate", SampleRate);
    Ini->Write("SampleClockSource", SampleClockSource);
	Ini->Write("ReferenceSource", ReferenceSource);
	Ini->Write("ReferenceFrequency", ReferenceFrequency);
    Ini->Write("BusmasterSize", BusmasterSize);
    Ini->Write("ExternalTrigger", ExternalTrigger);
    Ini->Write("Framed", Framed);
    Ini->Write("FrameCount", FrameCount);
    Ini->Write("AutoTrigger", AutoTrigger);
    Ini->Write("ScriptEnable", ScriptEnable);
    Ini->Write("BssScript", BssScript);
    Ini->Write("AssScript", AssScript);
    Ini->Write("DebugScript", DebugScript);
    Ini->Write("PhaseDetectorFrequency", PhaseDetectorFrequency);
    Ini->Write("WaveFile", WaveFile);
    Ini->Write("WaveType", WaveType);
    Ini->Write("WaveformFrequency", WaveformFrequency);
    Ini->Write("WaveformAmplitude", WaveformAmplitude);
	    Ini->Write("DioConfig", DioConfig);
    Ini->Write("FrontPanelConfig", FrontPanelConfig);
    Ini->Write("PllCorrection", PllCorrection);
    Ini->Write("PllCalibrated", PllCalibrated);

	Ini->Write("StreamPacketSize", StreamPacketSize);

    Ini->Write("PrefillPeriod", PrefillPeriod);

    Ini->Write("ExoFile", ExoFile);
    
 	Ini->Write("PllCorrection", PllCorrection);
	Ini->Write("PllCalibrated", PllCalibrated);

	Ini->Write("TestModeEnable", TestModeEnable);
	Ini->Write("FastClock", FastClock);

	delete Ini;
    Ini = 0;
}
